from bisect import bisect
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import overload
import threading

from ptychodus.api.observer import ObservableSequence


@dataclass(frozen=True)
class GlobusStatus:
    label: str
    start_time: datetime
    completion_time: datetime | None
    status: str
    action: str
    run_id: str

    @property
    def run_url(self) -> str:
        return f'https://app.globus.org/runs/{self.run_id}/logs'


class GlobusStatusRepository(ObservableSequence[GlobusStatus]):
    def __init__(self) -> None:
        super().__init__()
        self._status_list: list[GlobusStatus] = list()
        self._refresh_status_event = threading.Event()

    @overload
    def __getitem__(self, index: int) -> GlobusStatus: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[GlobusStatus]: ...

    def __getitem__(self, index: int | slice) -> GlobusStatus | Sequence[GlobusStatus]:
        return self._status_list[index]

    def __len__(self) -> int:
        return len(self._status_list)

    def needs_status_refresh(self) -> bool:
        return self._refresh_status_event.is_set()

    def refresh_status(self) -> None:
        self._refresh_status_event.set()

    def _status_refreshed(self) -> None:
        self._refresh_status_event.clear()

    def _insert_sorted(self, status: GlobusStatus) -> None:
        index = bisect(self._status_list, status.start_time, key=lambda x: x.start_time)
        self._status_list.insert(index, status)
        self.notify_observers_item_inserted(index, status)

    def _pop(self, index: int) -> GlobusStatus:
        status = self._status_list.pop(index)
        self.notify_observers_item_removed(index, status)
        return status

    def _update_if_changed(self, index: int, status: GlobusStatus) -> None:
        old_status = self._status_list[index]

        if old_status != status:
            self._status_list[index] = status
            self.notify_observers_item_changed(index, status)


class UpdateGlobusStatusRepository:
    def __init__(
        self, status_repository: GlobusStatusRepository, status_sequence: Sequence[GlobusStatus]
    ) -> None:
        self._status_repository = status_repository
        self._new_status_list = sorted(status_sequence, key=lambda x: x.start_time)

    def __call__(self) -> None:
        run_id_to_new_status_map = {status.run_id: status for status in self._new_status_list}

        for index in reversed(range(len(self._status_repository))):
            old_status = self._status_repository[index]

            try:
                new_status = run_id_to_new_status_map.pop(old_status.run_id)
            except KeyError:
                self._status_repository._pop(index)
            else:
                self._status_repository._update_if_changed(index, new_status)

        for new_status in run_id_to_new_status_map.values():
            self._status_repository._insert_sorted(new_status)

        self._status_repository._status_refreshed()



from __future__ import annotations
from collections.abc import Iterator

import numpy
import numpy.typing

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.probe import FresnelZonePlate, ProbeSequence, ProbeGeometryProvider
from ptychodus.api.propagator import FresnelTransformPropagator, PropagatorParameters

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class FresnelZonePlateProbeBuilder(ProbeSequenceBuilder):
    def __init__(
        self,
        settings: ProbeSettings,
        fresnel_zone_plate_chooser: PluginChooser[FresnelZonePlate],
    ) -> None:
        super().__init__(settings, 'fresnel_zone_plate')
        self._settings = settings
        self._fresnel_zone_plate_chooser = fresnel_zone_plate_chooser

        self.zone_plate_diameter_m = settings.zone_plate_diameter_m.copy()
        self._add_parameter('zone_plate_diameter_m', self.zone_plate_diameter_m)

        self.outermost_zone_width_m = settings.outermost_zone_width_m.copy()
        self._add_parameter('outermost_zone_width_m', self.outermost_zone_width_m)

        self.central_beamstop_diameter_m = settings.central_beamstop_diameter_m.copy()
        self._add_parameter('central_beamstop_diameter_m', self.central_beamstop_diameter_m)

        # from sample to the focal plane
        self.defocus_distance_m = settings.defocus_distance_m.copy()
        self._add_parameter('defocus_distance_m', self.defocus_distance_m)

    def copy(self) -> FresnelZonePlateProbeBuilder:
        builder = FresnelZonePlateProbeBuilder(self._settings, self._fresnel_zone_plate_chooser)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def labels_for_presets(self) -> Iterator[str]:
        for plugin in self._fresnel_zone_plate_chooser:
            yield plugin.display_name

    def apply_presets(self, display_name: str) -> None:
        self._fresnel_zone_plate_chooser.set_current_plugin(display_name)
        fzp = self._fresnel_zone_plate_chooser.get_current_plugin().strategy
        self.zone_plate_diameter_m.set_value(fzp.zone_plate_diameter_m)
        self.outermost_zone_width_m.set_value(fzp.outermost_zone_width_m)
        self.central_beamstop_diameter_m.set_value(fzp.central_beamstop_diameter_m)

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        wavelength_m = geometry_provider.probe_wavelength_m
        zone_plate = FresnelZonePlate(
            zone_plate_diameter_m=self.zone_plate_diameter_m.get_value(),
            outermost_zone_width_m=self.outermost_zone_width_m.get_value(),
            central_beamstop_diameter_m=self.central_beamstop_diameter_m.get_value(),
        )
        focal_length_m = zone_plate.get_focal_length_m(wavelength_m)
        distance_m = focal_length_m + self.defocus_distance_m.get_value()
        sample_plane_geometry = geometry_provider.get_probe_geometry()
        fzp_half_width = (sample_plane_geometry.width_px + 1) // 2
        fzp_half_height = (sample_plane_geometry.height_px + 1) // 2
        fzp_plane_pixel_size_numerator = wavelength_m * distance_m
        fzp_pixel_geometry = PixelGeometry(
            width_m=fzp_plane_pixel_size_numerator / sample_plane_geometry.width_m,
            height_m=fzp_plane_pixel_size_numerator / sample_plane_geometry.height_m,
        )

        # coordinate on FZP plane
        lx_fzp = -fzp_pixel_geometry.width_m * numpy.arange(-fzp_half_width, fzp_half_width)
        ly_fzp = -fzp_pixel_geometry.height_m * numpy.arange(-fzp_half_height, fzp_half_height)

        XX_FZP, YY_FZP = numpy.meshgrid(lx_fzp, ly_fzp)  # noqa: N806
        RR_FZP = numpy.hypot(XX_FZP, YY_FZP)  # noqa: N806

        # transmission function of FZP
        T = numpy.exp(  # noqa: N806
            -2j * numpy.pi / wavelength_m * (XX_FZP**2 + YY_FZP**2) / 2 / focal_length_m
        )
        C = RR_FZP <= zone_plate.zone_plate_diameter_m / 2  # noqa: N806
        H = RR_FZP >= zone_plate.central_beamstop_diameter_m / 2  # noqa: N806
        fzp_transmission_function = T * C * H

        propagator_parameters = PropagatorParameters(
            wavelength_m=wavelength_m,
            width_px=fzp_transmission_function.shape[-1],
            height_px=fzp_transmission_function.shape[-2],
            pixel_width_m=fzp_pixel_geometry.width_m,
            pixel_height_m=fzp_pixel_geometry.height_m,
            propagation_distance_m=distance_m,
        )
        propagator = FresnelTransformPropagator(propagator_parameters)
        array = propagator.propagate(fzp_transmission_function)

        return ProbeSequence(
            array=self.normalize(array),
            opr_weights=None,
            pixel_geometry=sample_plane_geometry.get_pixel_geometry(),
        )



from dataclasses import dataclass

import psutil


@dataclass(frozen=True)
class MemoryStatistics:
    total_physical_memory_bytes: int
    available_memory_bytes: int
    percent_usage: float


class MemoryPresenter:
    def get_statistics(self) -> MemoryStatistics:
        mem = psutil.virtual_memory()
        return MemoryStatistics(
            total_physical_memory_bytes=mem.total,
            available_memory_bytes=mem.available,
            percent_usage=mem.percent,
        )



from collections.abc import Sequence
from dataclasses import dataclass

import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.geometry import AffineTransform
from ptychodus.api.observer import Observable

from ..product import ProbePositionsRepository
from .settings import AffineTransformEstimatorSettings

__all__ = ['AffineTransformEstimator']


@dataclass(frozen=True)
class PreprocessedCoordinates:
    coordinates: RealArrayType
    centroid_x: float
    centroid_y: float
    rms_distance: float


def estimate_mean_hodges_lehman(values: RealArrayType) -> float:
    mean = numpy.median((values[numpy.newaxis, :] + values[:, numpy.newaxis]) / 2)
    return float(mean)


def estimate_affine_transform(
    uncorrected_coordinates: RealArrayType,
    corrected_coordinates: RealArrayType,
) -> AffineTransform:
    ones_col = numpy.ones((uncorrected_coordinates.shape[0]))
    a = numpy.repeat(numpy.column_stack((uncorrected_coordinates, ones_col)), 2, axis=0)
    b = corrected_coordinates.flatten()
    x, residuals, rank, singular_values = numpy.linalg.lstsq(a, b)
    return AffineTransform(x[0], x[1], x[2], x[3], x[4], x[5])


def evaluate_error(
    uncorrected_coordinates: RealArrayType,
    corrected_coordinates: RealArrayType,
    model: AffineTransform,
) -> RealArrayType:
    y0 = uncorrected_coordinates[-2]
    x0 = uncorrected_coordinates[-1]

    transform = numpy.vectorize(model.__call__)
    yt, xt = transform(y0, x0)

    y1 = corrected_coordinates[-2]
    x1 = corrected_coordinates[-1]

    return numpy.hypot(xt - x1, yt - y1)


class AffineTransformEstimator(Observable):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: AffineTransformEstimatorSettings,
        repository: ProbePositionsRepository,
    ) -> None:
        self._rng = rng
        self._settings = settings
        self._repository = repository

    def _preprocess_coordinates(self, product_indexes: Sequence[int]) -> PreprocessedCoordinates:
        coordinate_list: list[float] = []

        for product_index in product_indexes:
            positions = self._repository[product_index].get_probe_positions()

            for point in positions:
                coordinate_list.append(point.coordinate_x_m)
                coordinate_list.append(point.coordinate_y_m)

        coordinates = numpy.reshape(coordinate_list, (-1, 2))

        # robust centroid estimation
        centroid_x = estimate_mean_hodges_lehman(coordinates[:, -1])
        centroid_y = estimate_mean_hodges_lehman(coordinates[:, -2])
        coordinates -= numpy.array((centroid_x, centroid_y))

        # rescale for RMS distance = 1
        distance = numpy.hypot(coordinates[:, -1], coordinates[:, -2])
        rms_distance = numpy.sqrt(numpy.mean(numpy.square(distance)))
        coordinates /= rms_distance

        return PreprocessedCoordinates(coordinates, centroid_x, centroid_y, rms_distance)

    def estimate(
        self,
        measured_product_indexes: Sequence[int],
        corrected_product_indexes: Sequence[int],
    ) -> AffineTransform:
        corrected_set = set(corrected_product_indexes)
        measured_set = set(measured_product_indexes)

        if len(corrected_set) != len(corrected_product_indexes):
            raise ValueError('One or more duplicated corrected product indexes!')

        if len(measured_set) != len(measured_product_indexes):
            raise ValueError('One or more duplicated measured product indexes!')

        if not corrected_set.isdisjoint(measured_set):
            raise ValueError('Product index appears in corrected and measured sets!')

        corrected_coordinates = self._preprocess_coordinates(corrected_product_indexes)
        measured_coordinates = self._preprocess_coordinates(measured_product_indexes)
        indexes = numpy.arange(measured_coordinates.coordinates.shape[0])
        num_shuffles = self._settings.num_shuffles.get_value()
        inlier_threshold = self._settings.inlier_threshold.get_value()
        min_inliers = self._settings.min_inliers.get_value()
        arity = 3  # minimum number of points needed to estimate the model

        best_error = numpy.inf
        best_model = AffineTransform(1.0, 0.0, 0.0, 0.0, 1.0, 0.0)

        # RANSAC estimation of affine transform
        for it in range(num_shuffles):
            self._rng.shuffle(indexes)

            for chunk in range(0, len(indexes), arity):
                samples = indexes[chunk : chunk + arity]

                corrected_subset = numpy.take(corrected_coordinates.coordinates, samples, axis=0)
                uncorrected_subset = numpy.take(measured_coordinates.coordinates, samples, axis=0)
                coarse_model = estimate_affine_transform(uncorrected_subset, corrected_subset)
                error = evaluate_error(uncorrected_subset, corrected_subset, coarse_model)
                inliers = numpy.where(error < inlier_threshold)

                if len(inliers) > min_inliers:
                    corrected_subset = numpy.take(
                        corrected_coordinates.coordinates, inliers, axis=0
                    )
                    uncorrected_subset = numpy.take(
                        measured_coordinates.coordinates, inliers, axis=0
                    )
                    candidate_model = estimate_affine_transform(
                        uncorrected_subset, corrected_subset
                    )
                    candidate_error = evaluate_error(
                        uncorrected_subset, corrected_subset, candidate_model
                    )
                    candidate_error_rms = numpy.sqrt(numpy.mean(numpy.square(candidate_error)))

                    if candidate_error < best_error:
                        best_error = candidate_error_rms
                        best_model = candidate_model

        # TODO broken: unscale best_model

        return best_model



from collections.abc import Sequence

from ptychodus.api.reconstructor import ReconstructorLibrary
from ptychodus.api.settings import SettingsRegistry

from ..diffraction import DiffractionAPI
from ..product import ProductAPI
from ..task_manager import TaskManager
from .api import ProcessingAPI, ProcessingAlgorithmParameter
from .context import ProcessingContext
from .settings import ProcessingSettings


class ProcessingCore:
    def __init__(
        self,
        task_manager: TaskManager,
        settings_registry: SettingsRegistry,
        diffraction_api: DiffractionAPI,
        product_api: ProductAPI,
        algorithm_libraries: Sequence[ReconstructorLibrary],
    ) -> None:
        self._settings = ProcessingSettings(settings_registry)
        self.algorithm_parameter = ProcessingAlgorithmParameter(
            self._settings.algorithm, algorithm_libraries
        )
        self._context = ProcessingContext(task_manager)
        self.processing_api = ProcessingAPI(
            task_manager,
            diffraction_api,
            product_api,
            self.algorithm_parameter,
            self._context,
        )

        for library in algorithm_libraries:
            library.get_logger().addHandler(self._context.get_log_handler())



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class GlobusJob:
    flow_input: Mapping[str, Any]
    label: str
    tags: Sequence[str]


class GlobusClient(ABC):
    @property
    @abstractmethod
    def is_supported(self) -> bool:
        pass

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass

    @abstractmethod
    def run_flow(self, job: GlobusJob) -> None:
        pass


class FakeGlobusClient(GlobusClient):
    @property
    def is_supported(self) -> bool:
        return False

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    def run_flow(self, job: GlobusJob) -> None:
        pass



from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class PtyChiSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChi')
        self._group.add_observer(self)

        # ReconstructorOptions
        self.num_epochs = self._group.create_integer_parameter('NumEpochs', 100, minimum=1)
        self.num_sync_epochs = self._group.create_integer_parameter('NumSyncEpochs', 1, minimum=1)
        self.batch_size = self._group.create_integer_parameter('BatchSize', 100, minimum=1)
        self.batching_mode = self._group.create_string_parameter('BatchingMode', 'random')
        self.compact_mode_update_clustering = self._group.create_integer_parameter(
            'CompactModeUpdateClustering', 1, minimum=0
        )
        self.use_devices = self._group.create_boolean_parameter('UseDevices', True)
        self.use_double_precision = self._group.create_boolean_parameter(
            'UseDoublePrecision', False
        )
        self.use_double_precision_for_fft = self._group.create_boolean_parameter(
            'UseDoublePrecisionForFFT', False
        )
        self.allow_nondeterministic_algorithms = self._group.create_boolean_parameter(
            'AllowNondeterministicAlgorithms', True
        )

        # ForwardModelOptions
        self.use_low_memory_mode = self._group.create_boolean_parameter('UseLowMemoryMode', False)
        self.pad_for_shift = self._group.create_integer_parameter('PadForShift', 0, minimum=0)

        # PtychographyDataOptions
        self.use_far_field_propagation = self._group.create_boolean_parameter(
            'UseFarFieldPropagation', True
        )
        self.fft_shift_diffraction_patterns = self._group.create_boolean_parameter(
            'FFTShiftDiffractionPatterns', True
        )
        self.save_data_on_device = self._group.create_boolean_parameter('SaveDataOnDevice', False)

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiObjectSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiObject')
        self._group.add_observer(self)

        self.is_optimizable = self._group.create_boolean_parameter('IsOptimizable', True)
        self.optimization_plan_start = self._group.create_integer_parameter(
            'OptimizationPlanStart', 0, minimum=0
        )
        self.optimization_plan_stop = self._group.create_integer_parameter(
            'OptimizationPlanStop', -1
        )
        self.optimization_plan_stride = self._group.create_integer_parameter(
            'OptimizationPlanStride', 1, minimum=1
        )
        self.optimizer = self._group.create_string_parameter('Optimizer', 'SGD')
        self.step_size = self._group.create_real_parameter('StepSize', 1.0, minimum=0.0)

        self.optimize_slice_spacing = self._group.create_boolean_parameter(
            'OptimizeSliceSpacing', False
        )
        self.optimize_slice_spacing_start = self._group.create_integer_parameter(
            'OptimizeSliceSpacingStart', 0, minimum=0
        )
        self.optimize_slice_spacing_stop = self._group.create_integer_parameter(
            'OptimizeSliceSpacingStop', -1
        )
        self.optimize_slice_spacing_stride = self._group.create_integer_parameter(
            'OptimizeSliceSpacingStride', 1, minimum=1
        )
        self.optimize_slice_spacing_optimizer = self._group.create_string_parameter(
            'OptimizeSliceSpacingOptimizer', 'SGD'
        )
        self.optimize_slice_spacing_step_size = self._group.create_real_parameter(
            'OptimizeSliceSpacingStepSize', 1.0e-10, minimum=0.0
        )

        self.constrain_l1_norm = self._group.create_boolean_parameter('ConstrainL1Norm', False)
        self.constrain_l1_norm_start = self._group.create_integer_parameter(
            'ConstrainL1NormStart', 0, minimum=0
        )
        self.constrain_l1_norm_stop = self._group.create_integer_parameter(
            'ConstrainL1NormStop', -1
        )
        self.constrain_l1_norm_stride = self._group.create_integer_parameter(
            'ConstrainL1NormStride', 1, minimum=1
        )
        self.constrain_l1_norm_weight = self._group.create_real_parameter(
            'ConstrainL1NormWeight', 0.0, minimum=0.0
        )

        self.constrain_l2_norm = self._group.create_boolean_parameter('ConstrainL2Norm', False)
        self.constrain_l2_norm_start = self._group.create_integer_parameter(
            'ConstrainL2NormStart', 0, minimum=0
        )
        self.constrain_l2_norm_stop = self._group.create_integer_parameter(
            'ConstrainL2NormStop', -1
        )
        self.constrain_l2_norm_stride = self._group.create_integer_parameter(
            'ConstrainL2NormStride', 1, minimum=1
        )
        self.constrain_l2_norm_weight = self._group.create_real_parameter(
            'ConstrainL2NormWeight', 0.0, minimum=0.0
        )

        self.constrain_smoothness = self._group.create_boolean_parameter(
            'ConstrainSmoothness', False
        )
        self.constrain_smoothness_start = self._group.create_integer_parameter(
            'ConstrainSmoothnessStart', 0, minimum=0
        )
        self.constrain_smoothness_stop = self._group.create_integer_parameter(
            'ConstrainSmoothnessStop', -1
        )
        self.constrain_smoothness_stride = self._group.create_integer_parameter(
            'ConstrainSmoothnessStride', 1, minimum=1
        )
        self.constrain_smoothness_alpha = self._group.create_real_parameter(
            'ConstrainSmoothnessAlpha', 0.0, minimum=0.0, maximum=1.0 / 8
        )

        self.constrain_total_variation = self._group.create_boolean_parameter(
            'ConstrainTotalVariation', False
        )
        self.constrain_total_variation_start = self._group.create_integer_parameter(
            'ConstrainTotalVariationStart', 0, minimum=0
        )
        self.constrain_total_variation_stop = self._group.create_integer_parameter(
            'ConstrainTotalVariationStop', -1
        )
        self.constrain_total_variation_stride = self._group.create_integer_parameter(
            'ConstrainTotalVariationStride', 1, minimum=1
        )
        self.constrain_total_variation_weight = self._group.create_real_parameter(
            'ConstrainTotalVariationWeight', 0.0, minimum=0.0
        )

        self.remove_grid_artifacts = self._group.create_boolean_parameter(
            'RemoveGridArtifacts', False
        )
        self.remove_grid_artifacts_start = self._group.create_integer_parameter(
            'RemoveGridArtifactsStart', 0, minimum=0
        )
        self.remove_grid_artifacts_stop = self._group.create_integer_parameter(
            'RemoveGridArtifactsStop', -1
        )
        self.remove_grid_artifacts_stride = self._group.create_integer_parameter(
            'RemoveGridArtifactsStride', 1, minimum=1
        )
        self.remove_grid_artifacts_period_x_m = self._group.create_real_parameter(
            'RemoveGridArtifactsPeriodXInMeters', 1e-7, minimum=0.0
        )
        self.remove_grid_artifacts_period_y_m = self._group.create_real_parameter(
            'RemoveGridArtifactsPeriodYInMeters', 1e-7, minimum=0.0
        )
        self.remove_grid_artifacts_window_size_px = self._group.create_integer_parameter(
            'RemoveGridArtifactsWindowSizeInPixels',
            5,
            minimum=1,
        )
        self.remove_grid_artifacts_direction = self._group.create_string_parameter(
            'RemoveGridArtifactsDirection', 'XY'
        )

        self.regularize_multislice = self._group.create_boolean_parameter(
            'RegularizeMultislice', False
        )
        self.regularize_multislice_start = self._group.create_integer_parameter(
            'RegularizeMultisliceStart', 0, minimum=0
        )
        self.regularize_multislice_stop = self._group.create_integer_parameter(
            'RegularizeMultisliceStop', -1
        )
        self.regularize_multislice_stride = self._group.create_integer_parameter(
            'RegularizeMultisliceStride', 1, minimum=1
        )
        self.regularize_multislice_weight = self._group.create_real_parameter(
            'RegularizeMultisliceWeight', 0.0, minimum=0.0
        )
        self.regularize_multislice_unwrap_phase = self._group.create_boolean_parameter(
            'RegularizeMultisliceUnwrapPhase', True
        )
        self.regularize_multislice_unwrap_phase_image_gradient_method = (
            self._group.create_string_parameter(
                'RegularizeMultisliceUnwrapPhaseImageGradientMethod', 'FOURIER_DIFFERENTIATION'
            )
        )
        self.regularize_multislice_unwrap_phase_image_integration_method = (
            self._group.create_string_parameter(
                'RegularizeMultisliceUnwrapPhaseImageIntegrationMethod', 'FOURIER'
            )
        )

        self.patch_interpolator = self._group.create_string_parameter(
            'PatchInterpolator', 'FOURIER'
        )

        self.remove_object_probe_ambiguity = self._group.create_boolean_parameter(
            'RemoveObjectProbeAmbiguity', True
        )
        self.remove_object_probe_ambiguity_start = self._group.create_integer_parameter(
            'RemoveObjectProbeAmbiguityStart', 0, minimum=0
        )
        self.remove_object_probe_ambiguity_stop = self._group.create_integer_parameter(
            'RemoveObjectProbeAmbiguityStop', -1
        )
        self.remove_object_probe_ambiguity_stride = self._group.create_integer_parameter(
            'RemoveObjectProbeAmbiguityStride', 10, minimum=1
        )

        self.build_preconditioner_with_all_modes = self._group.create_boolean_parameter(
            'BuildPreconditionerWithAllModes', False
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiProbeSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiProbe')
        self._group.add_observer(self)

        self.is_optimizable = self._group.create_boolean_parameter('IsOptimizable', True)
        self.optimization_plan_start = self._group.create_integer_parameter(
            'OptimizationPlanStart', 0, minimum=0
        )
        self.optimization_plan_stop = self._group.create_integer_parameter(
            'OptimizationPlanStop', -1
        )
        self.optimization_plan_stride = self._group.create_integer_parameter(
            'OptimizationPlanStride', 1, minimum=1
        )
        self.optimizer = self._group.create_string_parameter('Optimizer', 'SGD')
        self.step_size = self._group.create_real_parameter('StepSize', 1.0, minimum=0.0)

        self.constrain_probe_power = self._group.create_boolean_parameter(
            'ConstrainProbePower', False
        )
        self.constrain_probe_power_start = self._group.create_integer_parameter(
            'ConstrainProbePowerStart', 0, minimum=0
        )
        self.constrain_probe_power_stop = self._group.create_integer_parameter(
            'ConstrainProbePowerStop', -1
        )
        self.constrain_probe_power_stride = self._group.create_integer_parameter(
            'ConstrainProbePowerStride', 1, minimum=1
        )

        self.orthogonalize_incoherent_modes = self._group.create_boolean_parameter(
            'OrthogonalizeIncoherentModes', True
        )
        self.orthogonalize_incoherent_modes_start = self._group.create_integer_parameter(
            'OrthogonalizeIncoherentModesStart', 0, minimum=0
        )
        self.orthogonalize_incoherent_modes_stop = self._group.create_integer_parameter(
            'OrthogonalizeIncoherentModesStop', -1
        )
        self.orthogonalize_incoherent_modes_stride = self._group.create_integer_parameter(
            'OrthogonalizeIncoherentModesStride', 1, minimum=1
        )
        self.orthogonalize_incoherent_modes_method = self._group.create_string_parameter(
            'OrthogonalizeIncoherentModesMethod', 'SVD'
        )

        self.orthogonalize_opr_modes = self._group.create_boolean_parameter(
            'OrthogonalizeOPRModes', True
        )
        self.orthogonalize_opr_modes_start = self._group.create_integer_parameter(
            'OrthogonalizeOPRModesStart', 0, minimum=0
        )
        self.orthogonalize_opr_modes_stop = self._group.create_integer_parameter(
            'OrthogonalizeOPRModesStop', -1
        )
        self.orthogonalize_opr_modes_stride = self._group.create_integer_parameter(
            'OrthogonalizeOPRModesStride', 1, minimum=1
        )

        self.constrain_support = self._group.create_boolean_parameter('ConstrainSupport', False)
        self.constrain_support_start = self._group.create_integer_parameter(
            'ConstrainSupportStart', 0, minimum=0
        )
        self.constrain_support_stop = self._group.create_integer_parameter(
            'ConstrainSupportStop', -1
        )
        self.constrain_support_stride = self._group.create_integer_parameter(
            'ConstrainSupportStride', 1, minimum=1
        )
        self.constrain_support_threshold = self._group.create_real_parameter(
            'ConstrainSupportThreshold', 0.005, minimum=0.0
        )

        self.constrain_center = self._group.create_boolean_parameter('ConstrainCenter', False)
        self.constrain_center_start = self._group.create_integer_parameter(
            'ConstrainCenterStart', 0, minimum=0
        )
        self.constrain_center_stop = self._group.create_integer_parameter('ConstrainCenterStop', -1)
        self.constrain_center_stride = self._group.create_integer_parameter(
            'ConstrainCenterStride', 1, minimum=1
        )
        self.use_intensity_for_mass_centroid = self._group.create_boolean_parameter(
            'UseIntensityForMassCentroid', False
        )

        self.relax_eigenmode_update = self._group.create_real_parameter(
            'RelaxEigenmodeUpdate', 1.0, minimum=0.0, maximum=1.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiProbePositionSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiProbePosition')
        self._group.add_observer(self)

        self.is_optimizable = self._group.create_boolean_parameter('IsOptimizable', False)
        self.optimization_plan_start = self._group.create_integer_parameter(
            'OptimizationPlanStart', 0, minimum=0
        )
        self.optimization_plan_stop = self._group.create_integer_parameter(
            'OptimizationPlanStop', -1
        )
        self.optimization_plan_stride = self._group.create_integer_parameter(
            'OptimizationPlanStride', 1, minimum=1
        )
        self.optimizer = self._group.create_string_parameter('Optimizer', 'SGD')
        self.step_size = self._group.create_real_parameter('StepSize', 0.3, minimum=0.0)

        self.constrain_centroid = self._group.create_boolean_parameter('ConstrainCentroid', False)

        # correction_options
        self.correction_type = self._group.create_string_parameter('CorrectionType', 'GRADIENT')
        self.differentiation_method = self._group.create_string_parameter(
            'DifferentiationMethod', 'FOURIER_DIFFERENTIATION'
        )
        self.cross_correlation_scale = self._group.create_integer_parameter(
            'CrossCorrelationScale', 20000, minimum=1
        )
        self.cross_correlation_real_space_width = self._group.create_real_parameter(
            'CrossCorrelationRealSpaceWidth', 0.01, minimum=0.0
        )
        self.cross_correlation_probe_threshold = self._group.create_real_parameter(
            'CrossCorrelationProbeThreshold', 0.1, minimum=0.0, maximum=1.0
        )
        self.choose_slice_for_correction = self._group.create_boolean_parameter(
            'ChooseSliceForCorrection', False
        )
        self.slice_for_correction = self._group.create_integer_parameter(
            'SliceForCorrection', 0, minimum=0
        )
        self.clip_update_magnitude_by_mad = self._group.create_boolean_parameter(
            'ClipUpdateMagnitudeByMAD', True
        )
        self.limit_update_magnitude = self._group.create_boolean_parameter(
            'LimitUpdateMagnitude', False
        )
        self.update_magnitude_limit = self._group.create_real_parameter(
            'UpdateMagnitudeLimit', 0.1, minimum=0.0
        )

        # affine_transform_constraint
        self.constrain_affine_transform = self._group.create_boolean_parameter(
            'ConstrainAffineTransform', False
        )
        self.constrain_affine_transform_start = self._group.create_integer_parameter(
            'ConstrainAffineTransformStart', 0, minimum=0
        )
        self.constrain_affine_transform_stop = self._group.create_integer_parameter(
            'ConstrainAffineTransformStop', -1
        )
        self.constrain_affine_transform_stride = self._group.create_integer_parameter(
            'ConstrainAffineTransformStride', 1, minimum=1
        )
        self.constrain_affine_transform_degrees_of_freedom = self._group.create_integer_parameter(
            'ConstrainAffineTransformDegreesOfFreedom', 0, minimum=0
        )
        self.constrain_affine_transform_position_weight_update_interval = (
            self._group.create_integer_parameter(
                'ConstrainAffineTransformPositionWeightUpdateInterval', 10, minimum=1
            )
        )
        self.constrain_affine_transform_apply_constraint = self._group.create_boolean_parameter(
            'ConstrainAffineTransformApplyConstraint', True
        )
        self.constrain_affine_transform_max_expected_error_px = self._group.create_real_parameter(
            'ConstrainAffineTransformMaxExpectedErrorInPixels', 1.0, minimum=0.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiOPRSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiOPR')
        self._group.add_observer(self)

        self.is_optimizable = self._group.create_boolean_parameter('IsOptimizable', False)
        self.optimization_plan_start = self._group.create_integer_parameter(
            'OptimizationPlanStart', 0, minimum=0
        )
        self.optimization_plan_stop = self._group.create_integer_parameter(
            'OptimizationPlanStop', -1
        )
        self.optimization_plan_stride = self._group.create_integer_parameter(
            'OptimizationPlanStride', 1, minimum=1
        )
        self.optimizer = self._group.create_string_parameter('Optimizer', 'SGD')
        self.step_size = self._group.create_real_parameter('StepSize', 1.0, minimum=0.0)

        self.optimize_eigenmode_weights = self._group.create_boolean_parameter(
            'OptimizeEigenmodeWeigts', True
        )
        self.optimize_intensities = self._group.create_boolean_parameter(
            'OptimizeIntensities', False
        )

        self.smooth_mode_weights = self._group.create_boolean_parameter('SmoothModeWeights', False)
        self.smooth_mode_weights_start = self._group.create_integer_parameter(
            'SmoothModeWeightsStart', 0, minimum=0
        )
        self.smooth_mode_weights_stop = self._group.create_integer_parameter(
            'SmoothModeWeightsStop', -1
        )
        self.smooth_mode_weights_stride = self._group.create_integer_parameter(
            'SmoothModeWeightsStride', 1, minimum=1
        )
        self.smoothing_method = self._group.create_string_parameter('SmoothingMethod', '')
        self.polynomial_smoothing_degree = self._group.create_integer_parameter(
            'PolynomialSmoothingDegree', 4, minimum=0, maximum=10
        )

        self.relax_update = self._group.create_real_parameter(
            'RelaxUpdate', 1.0, minimum=0.0, maximum=1.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiAutodiffSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiAutodiff')
        self._group.add_observer(self)

        self.loss_function = self._group.create_string_parameter('LossFunction', 'MSE_SQRT')
        self.forward_model_class = self._group.create_string_parameter(
            'ForwardModelClass', 'PLANAR_PTYCHOGRAPHY'
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiDMSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiDM')
        self._group.add_observer(self)

        self.exit_wave_update_relaxation = self._group.create_real_parameter(
            'ExitWaveUpdateRelaxation', 1.0, minimum=0.0, maximum=1.0
        )
        self.chunk_length = self._group.create_integer_parameter('ChunkLength', 1, minimum=1)
        self.object_amplitude_clamp_limit = self._group.create_real_parameter(
            'ObjectAmplitudeClampLimit', 1000, minimum=0.0
        )
        self.object_inertia = self._group.create_real_parameter(
            'ObjectInertia', 0.0, minimum=0.0, maximum=1.0
        )
        self.probe_inertia = self._group.create_real_parameter(
            'ProbeInertia', 0.0, minimum=0.0, maximum=1.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiLSQMLSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiLSQML')
        self._group.add_observer(self)

        self.noise_model = self._group.create_string_parameter('NoiseModel', 'GAUSSIAN')
        self.gaussian_noise_deviation = self._group.create_real_parameter(
            'GaussianNoiseDeviation', 0.5
        )
        self.solve_object_probe_step_size_jointly_for_first_slice_in_multislice = (
            self._group.create_boolean_parameter(
                'SolveObjectProbeStepSizeJointlyForFirstSliceInMultislice', False
            )
        )
        self.solve_step_sizes_only_using_first_probe_mode = self._group.create_boolean_parameter(
            'SolveStepSizesOnlyUsingFirstProbeMode', True
        )
        self.momentum_acceleration_gain = self._group.create_real_parameter(
            'MomentumAccelerationGain', 0.0, minimum=0.0
        )
        self.use_momentum_acceleration_gradient_mixing_factor = (
            self._group.create_boolean_parameter(
                'UseMomentumAccelerationGradientMixingFactor', False
            )
        )
        self.momentum_acceleration_gradient_mixing_factor = self._group.create_real_parameter(
            'MomentumAccelerationGradientMixingFactor', 1.0
        )
        self.rescale_probe_intensity_in_first_epoch = self._group.create_boolean_parameter(
            'RescaleProbeIntensityInFirstEpoch', True
        )
        self.preconditioning_damping_factor = self._group.create_real_parameter(
            'PreconditioningDampingFactor', 0.1, minimum=0.0
        )

        self.object_optimal_step_size_scaler = self._group.create_real_parameter(
            'ObjectOptimalStepSizeScaler', 0.9, minimum=0.0
        )
        self.object_multimodal_update = self._group.create_boolean_parameter(
            'ObjectMultimodalUpdate', True
        )
        self.probe_optimal_step_size_scaler = self._group.create_real_parameter(
            'ProbeOptimalStepSizeScaler', 0.9, minimum=0.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtyChiPIESettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtyChiPIE')
        self._group.add_observer(self)

        self.probe_alpha = self._group.create_real_parameter(
            'ProbeAlpha', 0.1, minimum=0.0, maximum=1.0
        )
        self.object_alpha = self._group.create_real_parameter(
            'ObjectAlpha', 0.1, minimum=0.0, maximum=1.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.propagator import AngularSpectrumPropagator, PropagatorParameters

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class DiskProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, settings: ProbeSettings) -> None:
        super().__init__(settings, 'disk')
        self._settings = settings

        self.diameter_m = settings.disk_diameter_m.copy()
        self._add_parameter('diameter_m', self.diameter_m)

        # from sample to the focal plane
        self.defocus_distance_m = settings.defocus_distance_m.copy()
        self._add_parameter('defocus_distance_m', self.defocus_distance_m)

    def copy(self) -> DiskProbeBuilder:
        builder = DiskProbeBuilder(self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        geometry = geometry_provider.get_probe_geometry()
        coords = self.get_transverse_coordinates(geometry)

        R_m = coords.position_r_m  # noqa: N806
        r_m = self.diameter_m.get_value() / 2.0
        disk = numpy.where(R_m < r_m, 1 + 0j, 0j)

        propagator_parameters = PropagatorParameters(
            wavelength_m=geometry_provider.probe_wavelength_m,
            width_px=disk.shape[-1],
            height_px=disk.shape[-2],
            pixel_width_m=geometry.pixel_width_m,
            pixel_height_m=geometry.pixel_height_m,
            propagation_distance_m=self.defocus_distance_m.get_value(),
        )
        propagator = AngularSpectrumPropagator(propagator_parameters)
        array = propagator.propagate(disk)

        return ProbeSequence(
            array=self.normalize(array),
            opr_weights=None,
            pixel_geometry=geometry.get_pixel_geometry(),
        )



from __future__ import annotations
import logging


from ptychodus.api.object import Object, ObjectGeometryProvider
from ptychodus.api.observer import Observable
from ptychodus.api.parametric import ParameterGroup

from .builder import FromMemoryObjectBuilder, ObjectBuilder
from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class ObjectRepositoryItem(ParameterGroup):
    def __init__(
        self,
        geometry_provider: ObjectGeometryProvider,
        settings: ObjectSettings,
        builder: ObjectBuilder,
    ) -> None:
        super().__init__()
        self._geometry_provider = geometry_provider
        self._settings = settings
        self._builder = builder
        self._object = Object(array=None, pixel_geometry=None, center=None)

        self.layer_spacing_m = settings.object_layer_spacing_m.copy()
        self._add_parameter('layer_spacing_m', self.layer_spacing_m)

        self._add_group('builder', builder, observe=True)
        self.rebuild()

    def assign_item(self, item: ObjectRepositoryItem) -> None:
        self.layer_spacing_m.set_value(item.layer_spacing_m.get_value(), notify=False)
        self.set_builder(item.get_builder().copy())
        self.rebuild()

    def assign(self, object_: Object) -> None:
        builder = FromMemoryObjectBuilder(self._settings, object_)
        self.set_builder(builder)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

        self._builder.sync_to_settings()

    def get_num_layers(self) -> int:
        return len(self.layer_spacing_m) + 1

    def set_num_layers(self, num_layers: int) -> None:
        num_spaces = max(0, num_layers - 1)
        distance_m = list(self.layer_spacing_m.get_value())

        try:
            default_distance_m = distance_m[-1]
        except IndexError:
            default_distance_m = 0.0

        while len(distance_m) < num_spaces:
            distance_m.append(default_distance_m)

        if len(distance_m) > num_spaces:
            distance_m = distance_m[:num_spaces]

        self.layer_spacing_m.set_value(distance_m)
        self.rebuild()

    def get_object(self) -> Object:
        return self._object

    def get_builder(self) -> ObjectBuilder:
        return self._builder

    def set_builder(self, builder: ObjectBuilder) -> None:
        self._remove_group('builder')
        self._builder.remove_observer(self)
        self._builder = builder
        self._builder.add_observer(self)
        self._add_group('builder', self._builder, observe=True)
        self.rebuild()

    def rebuild(self, *, recenter: bool = False) -> None:
        try:
            object_ = self._builder.build(self._geometry_provider, self.layer_spacing_m.get_value())
        except Exception:
            logger.exception('Failed to rebuild object!')
            return

        if recenter:
            object_geometry = self._geometry_provider.get_object_geometry()
            self._object = Object(
                array=object_.get_array(),
                layer_spacing_m=object_.layer_spacing_m,
                pixel_geometry=object_.get_pixel_geometry(),
                center=object_geometry.get_center(),
            )
        else:
            self._object = object_

        self.layer_spacing_m.set_value(object_.layer_spacing_m)
        self.notify_observers()

    def _update(self, observable: Observable) -> None:
        if observable is self._builder:
            self.rebuild()
        else:
            super()._update(observable)



from __future__ import annotations
from collections.abc import Iterator, Mapping
from pathlib import Path
from typing import Any
import logging

from ptychodus.api.diffraction import CropCenter
from ptychodus.api.geometry import AffineTransform, ImageExtent
from ptychodus.api.product import Product
from ptychodus.api.reconstructor import AssembledDiffractionData, ReconstructInput
from ptychodus.api.settings import PathPrefixChange, SettingsRegistry
from ptychodus.api.workflow import WorkflowAPI, WorkflowDiffractionAPI, WorkflowProductAPI

from .diffraction import DiffractionAPI
from .globus import GlobusExecutor
from .product import ObjectAPI, ProbeAPI, ProductAPI, ProbePositionsAPI
from .processing import ProcessingAPI

logger = logging.getLogger(__name__)


class ConcreteWorkflowDiffractionAPI(WorkflowDiffractionAPI):
    def __init__(self, diffraction_api: DiffractionAPI) -> None:
        self._diffraction_api = diffraction_api

    def get_assembled_data(self) -> AssembledDiffractionData:
        return self._diffraction_api.get_assembled_data()

    def save_assembled_data(self, file_path: Path) -> None:
        self._diffraction_api.export_assembled_patterns(file_path)


class ConcreteWorkflowProductAPI(WorkflowProductAPI):
    def __init__(
        self,
        product_api: ProductAPI,
        probe_positions_api: ProbePositionsAPI,
        probe_api: ProbeAPI,
        object_api: ObjectAPI,
        processing_api: ProcessingAPI,
        globus_executor: GlobusExecutor,
        product_index: int,
    ) -> None:
        self._product_api = product_api
        self._probe_positions_api = probe_positions_api
        self._probe_api = probe_api
        self._object_api = object_api
        self._processing_api = processing_api
        self._globus_executor = globus_executor
        self._product_index = product_index

    def get_product_index(self) -> int:
        return self._product_index

    def get_product(self) -> Product:
        item = self._product_api.get_item(self._product_index)
        return item.get_product()

    def rename_product(self, new_name: str) -> None:
        self._product_api.rename_product(self._product_index, new_name)

    def load_probe_positions(self, file_path: Path, *, file_type: str | None = None) -> None:
        self._probe_positions_api.open_probe_positions(
            self._product_index, file_path, file_type=file_type
        )

    def generate_probe_positions(
        self,
        generator_name: str | None = None,
        generator_parameters: Mapping[str, Any] | None = None,
    ) -> None:
        if generator_name is None:
            self._probe_positions_api.build_probe_positions_from_settings(self._product_index)
        else:
            self._probe_positions_api.build_probe_positions(
                self._product_index, generator_name, generator_parameters
            )

    def set_probe_positions_transform(self, transform: AffineTransform) -> None:
        item = self._product_api.get_item(self._product_index)
        probe_positions_transform = item.get_probe_positions_item().get_transform()
        probe_positions_transform.set_transform(transform)

    def load_probe(self, file_path: Path, *, file_type: str | None = None) -> None:
        self._probe_api.open_probe(self._product_index, file_path, file_type=file_type)

    def generate_probe(
        self,
        generator_name: str | None = None,
        generator_parameters: Mapping[str, Any] | None = None,
    ) -> None:
        if generator_name is None:
            self._probe_api.build_probe_from_settings(self._product_index)
        else:
            self._probe_api.build_probe(self._product_index, generator_name, generator_parameters)

    def load_object(self, file_path: Path, *, file_type: str | None = None) -> None:
        self._object_api.open_object(self._product_index, file_path, file_type=file_type)

    def generate_object(
        self,
        generator_name: str | None = None,
        generator_parameters: Mapping[str, Any] | None = None,
    ) -> None:
        if generator_name is None:
            self._object_api.build_object_from_settings(self._product_index)
        else:
            self._object_api.build_object(self._product_index, generator_name, generator_parameters)

    def get_reconstruct_input(self) -> ReconstructInput:
        return self._processing_api.get_reconstruct_input(self._product_index)

    def reconstruct_local(
        self,
        *,
        algorithm: str | None = None,
        output_product_file: Path | None = None,
        block: bool = False,
    ) -> WorkflowProductAPI:
        output_product_index = self._processing_api.reconstruct(
            self._product_index,
            algorithm=algorithm,
            output_product_file=output_product_file,
            block=block,
        )

        return ConcreteWorkflowProductAPI(
            self._product_api,
            self._probe_positions_api,
            self._probe_api,
            self._object_api,
            self._processing_api,
            self._globus_executor,
            output_product_index,
        )

    def reconstruct_remote(self, *, algorithm: str | None = None) -> None:
        self._globus_executor.reconstruct(self._product_index, algorithm=algorithm)

    def train_reconstructor_local(
        self,
        input_path: Path,
        output_path: Path,
        *,
        algorithm: str | None = None,
        block: bool = False,
    ) -> None:
        # TODO mlflow
        self._processing_api.train(
            self._product_index, input_path, output_path, algorithm=algorithm, block=block
        )

    def train_reconstructor_remote(self, *, algorithm: str | None = None) -> None:
        # TODO mlflow
        self._globus_executor.train(self._product_index, algorithm=algorithm)

    def export_training_data(self, file_path: Path, *, algorithm: str | None = None) -> None:
        self._processing_api.export_training_data(
            file_path, self._product_index, algorithm=algorithm
        )

    def save_product(self, file_path: Path, *, file_type: str | None = None) -> None:
        self._product_api.save_product(self._product_index, file_path, file_type=file_type)


class ConcreteWorkflowAPI(WorkflowAPI):
    def __init__(
        self,
        settings_registry: SettingsRegistry,
        diffraction_api: DiffractionAPI,
        product_api: ProductAPI,
        probe_positions_api: ProbePositionsAPI,
        probe_api: ProbeAPI,
        object_api: ObjectAPI,
        processing_api: ProcessingAPI,
        executor: GlobusExecutor,
    ) -> None:
        self._settings_registry = settings_registry
        self._diffraction_api = diffraction_api
        self._product_api = product_api
        self._probe_positions_api = probe_positions_api
        self._probe_api = probe_api
        self._object_api = object_api
        self._processing_api = processing_api
        self._executor = executor

    def load_diffraction_data(
        self,
        file_path: Path,
        *,
        file_type: str | None = None,
        crop_center: CropCenter | None = None,
        crop_extent: ImageExtent | None = None,
        detector_extent: ImageExtent | None = None,
        process_patterns: bool = True,
        block: bool = False,
    ) -> WorkflowDiffractionAPI:
        self._diffraction_api.open_patterns(
            file_path,
            file_type=file_type,
            crop_center=crop_center,
            crop_extent=crop_extent,
            detector_extent=detector_extent,
            process_patterns=process_patterns,
            block=block,
        )
        return ConcreteWorkflowDiffractionAPI(self._diffraction_api)

    def available_reconstructors(self) -> Iterator[str]:
        return self._processing_api.available_reconstructors()

    def load_assembled_diffraction_data(self, file_path: Path) -> WorkflowDiffractionAPI:
        self._diffraction_api.import_assembled_patterns(file_path)
        return ConcreteWorkflowDiffractionAPI(self._diffraction_api)

    def get_product(self, product_index: int) -> WorkflowProductAPI:
        if product_index < 0:
            raise ValueError(f'Bad product index ({product_index=})!')

        return ConcreteWorkflowProductAPI(
            self._product_api,
            self._probe_positions_api,
            self._probe_api,
            self._object_api,
            self._processing_api,
            self._executor,
            product_index,
        )

    def register_product(self, product: Product) -> WorkflowProductAPI:
        product_index = self._product_api.insert_product(product)
        return self.get_product(product_index)

    def load_product(self, file_path: Path, *, file_type: str | None = None) -> WorkflowProductAPI:
        product_index = self._product_api.open_product(file_path, file_type=file_type)

        if product_index < 0:
            raise RuntimeError(f'Failed to open product "{file_path}"!')

        return self.get_product(product_index)

    def create_product(
        self,
        name: str,
        *,
        comments: str = '',
        detector_distance_m: float | None = None,
        probe_energy_eV: float | None = None,  # noqa: N803
        probe_photon_count: float | None = None,
        exposure_time_s: float | None = None,
        mass_attenuation_m2_kg: float | None = None,
        tomography_angle_deg: float | None = None,
    ) -> WorkflowProductAPI:
        product_index = self._product_api.insert_new_product(
            name,
            comments=comments,
            detector_distance_m=detector_distance_m,
            probe_energy_eV=probe_energy_eV,
            probe_photon_count=probe_photon_count,
            exposure_time_s=exposure_time_s,
            mass_attenuation_m2_kg=mass_attenuation_m2_kg,
            tomography_angle_deg=tomography_angle_deg,
        )
        return self.get_product(product_index)

    def save_settings(
        self, file_path: Path, change_path_prefix: PathPrefixChange | None = None
    ) -> None:
        self._settings_registry.save_settings(file_path, change_path_prefix)



from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Sequence
import logging

from ptychodus.api.observer import Observable
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.product import LossValue, Product

from .geometry import ProductGeometry
from .metadata import MetadataRepositoryItem, UniqueNameFactory
from .object import ObjectRepositoryItem
from .probe import ProbeRepositoryItem
from .probe_positions import ProbePositionsRepositoryItem
from .validator import ProductValidator

logger = logging.getLogger(__name__)


class ProductRepositoryItemObserver(UniqueNameFactory):
    @abstractmethod
    def handle_metadata_changed(self, item: ProductRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_probe_positions_changed(self, item: ProductRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_probe_changed(self, item: ProductRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_object_changed(self, item: ProductRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_losses_changed(self, item: ProductRepositoryItem) -> None:
        pass


class ProductRepositoryItem(ParameterGroup):
    def __init__(
        self,
        parent: ProductRepositoryItemObserver,
        metadata_item: MetadataRepositoryItem,
        probe_positions_item: ProbePositionsRepositoryItem,
        geometry: ProductGeometry,
        probe_item: ProbeRepositoryItem,
        object_item: ObjectRepositoryItem,
        validator: ProductValidator,
        losses: Sequence[LossValue],
    ) -> None:
        super().__init__()
        self._parent = parent
        self._metadata_item = metadata_item
        self._probe_positions_item = probe_positions_item
        self._geometry = geometry
        self._probe_item = probe_item
        self._object_item = object_item
        self._validator = validator
        self._losses = list(losses)

        self._add_group('metadata', self._metadata_item, observe=True)
        self._add_group('probe_positions', self._probe_positions_item, observe=True)
        self._add_group('probe', self._probe_item, observe=True)
        self._add_group('object', self._object_item, observe=True)

        self._index = -1  # used by ProductRepository

    def assign(self, product: Product) -> None:
        self._metadata_item.assign(product.metadata)
        self._probe_positions_item.assign(product.probe_positions)
        self._probe_item.assign(product.probes)
        self._object_item.assign(product.object_)
        self._losses = list(product.losses)
        self._parent.handle_losses_changed(self)

    def sync_to_settings(self) -> None:
        self._metadata_item.sync_to_settings()
        self._probe_positions_item.sync_to_settings()
        self._probe_item.sync_to_settings()
        self._object_item.sync_to_settings()

    def get_name(self) -> str:
        return self._metadata_item.name.get_value()

    def set_name(self, name: str) -> None:
        self._metadata_item.name.set_value(name)

    def get_metadata_item(self) -> MetadataRepositoryItem:
        return self._metadata_item

    def get_probe_positions_item(self) -> ProbePositionsRepositoryItem:
        return self._probe_positions_item

    def get_geometry(self) -> ProductGeometry:
        return self._geometry

    def get_probe_item(self) -> ProbeRepositoryItem:
        return self._probe_item

    def get_object_item(self) -> ObjectRepositoryItem:
        return self._object_item

    def _invalidate_losses(self) -> None:
        self._losses = list()
        self._parent.handle_losses_changed(self)

    def get_losses(self) -> Sequence[LossValue]:
        return self._losses

    def get_product(self) -> Product:
        return Product(
            metadata=self._metadata_item.get_metadata(),
            probe_positions=self._probe_positions_item.get_probe_positions(),
            probes=self._probe_item.get_probes(),
            object_=self._object_item.get_object(),
            losses=self.get_losses(),
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._metadata_item:
            self._invalidate_losses()
            self._parent.handle_metadata_changed(self)
        elif observable is self._probe_positions_item:
            self._invalidate_losses()
            self._parent.handle_probe_positions_changed(self)
        elif observable is self._probe_item:
            self._invalidate_losses()
            self._parent.handle_probe_changed(self)
        elif observable is self._object_item:
            self._invalidate_losses()
            self._parent.handle_object_changed(self)
        else:
            super()._update(observable)


class ProductRepositoryObserver(ABC):
    @abstractmethod
    def handle_item_inserted(self, index: int, item: ProductRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_metadata_changed(self, index: int, item: MetadataRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_probe_positions_changed(
        self, index: int, item: ProbePositionsRepositoryItem
    ) -> None:
        pass

    @abstractmethod
    def handle_probe_changed(self, index: int, item: ProbeRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_object_changed(self, index: int, item: ObjectRepositoryItem) -> None:
        pass

    @abstractmethod
    def handle_losses_changed(self, index: int, losses: Sequence[LossValue]) -> None:
        pass

    @abstractmethod
    def handle_item_removed(self, index: int, item: ProductRepositoryItem) -> None:
        pass



from collections.abc import Iterator, Sequence
import logging

from ptychodus.api.settings import SettingsRegistry

from .argo import ArgoChatTerminal
from .chat import ChatHistory, ChatTerminal
from .settings import ArgoSettings

logger = logging.getLogger(__name__)


class AgentPresenter:
    def __init__(self, terminal: ChatTerminal) -> None:
        self._terminal = terminal

    def get_available_chat_models(self) -> Iterator[str]:
        for model in [
            'gpt35',
            'gpt35large',
            'gpt4',
            'gpt4large',
            'gpt4o',
            'gpt4olatestgpt4turbo',
            'gpto1',
            'gpto1mini',
            'gpto3mini',
        ]:
            yield model

    def send_message(self, content: str) -> None:
        if self._terminal is not None:
            self._terminal.send_message(content)

    def get_available_embeddings_models(self) -> Iterator[str]:
        for model in ['ada002', 'v3large', 'v3small']:
            yield model

    def embed_text(self, texts: Sequence[str]) -> Sequence[Sequence[float]]:
        return [[]] if self._terminal is None else self._terminal.embed_texts(texts)


class AgentCore:
    def __init__(self, settings_registry: SettingsRegistry):
        self.settings = ArgoSettings(settings_registry)
        self.chat_history = ChatHistory()
        self._terminal = ArgoChatTerminal(self.settings, self.chat_history)
        self.presenter = AgentPresenter(self._terminal)



from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Final

import matplotlib.colors
import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import Parameter
from ptychodus.api.plugins import PluginChooser

__all__ = [
    'CylindricalColorModel',
    'CylindricalColorModelParameter',
]


def hsva_to_rgba(
    hue: RealArrayType, saturation: RealArrayType, value: RealArrayType, alpha: RealArrayType
) -> RealArrayType:
    hsv = numpy.stack((hue, saturation, value), axis=-1)
    rgb = matplotlib.colors.hsv_to_rgb(hsv)

    if alpha.ndim == 1:
        return numpy.column_stack((rgb, alpha))

    return numpy.dstack((rgb, alpha))


def _v(m1: RealArrayType, m2: RealArrayType, hue: RealArrayType) -> RealArrayType:
    """Adapted from colorsys._v in the Python standard library."""
    if m1.shape != hue.shape or m2.shape != hue.shape:
        raise ValueError('Shape mismatch: m1, m2, and hue must have the same shape.')

    hue = hue % 1.0

    return numpy.select(
        [hue < 1.0 / 6.0, hue < 0.5, hue < 2.0 / 3.0],
        [m1 + (m2 - m1) * hue * 6.0, m2, m1 + (m2 - m1) * (2.0 / 3.0 - hue) * 6.0],
        default=m1,
    )


def hlsa_to_rgba(
    hue: RealArrayType, lightness: RealArrayType, saturation: RealArrayType, alpha: RealArrayType
) -> RealArrayType:
    """Adapted from colorsys.hls_to_rgb in the Python standard library."""
    one_third: Final[float] = 1.0 / 3.0

    m2 = numpy.where(
        lightness <= 0.5,
        lightness * (1.0 + saturation),
        lightness + saturation - (lightness * saturation),
    )
    m1 = 2.0 * lightness - m2

    red = numpy.where(saturation > 0.0, _v(m1, m2, hue + one_third), lightness)
    green = numpy.where(saturation > 0.0, _v(m1, m2, hue), lightness)
    blue = numpy.where(saturation > 0.0, _v(m1, m2, hue - one_third), lightness)

    return numpy.stack((red, green, blue, alpha), axis=-1)


class CylindricalColorModel(ABC):
    @abstractmethod
    def __call__(self, h: RealArrayType, x: RealArrayType) -> RealArrayType:
        pass


class HSVSaturationColorModel(CylindricalColorModel):
    def __call__(self, h: RealArrayType, x: RealArrayType) -> RealArrayType:
        ones = numpy.ones_like(h)
        return hsva_to_rgba(h, x, ones, ones)


class HSVValueColorModel(CylindricalColorModel):
    def __call__(self, h: RealArrayType, x: RealArrayType) -> RealArrayType:
        ones = numpy.ones_like(h)
        return hsva_to_rgba(h, ones, x, ones)


class HSVAlphaColorModel(CylindricalColorModel):
    def __call__(self, h: RealArrayType, x: RealArrayType) -> RealArrayType:
        ones = numpy.ones_like(h)
        return hsva_to_rgba(h, ones, ones, x)


class HLSLightnessColorModel(CylindricalColorModel):
    def __call__(self, h: RealArrayType, x: RealArrayType) -> RealArrayType:
        ones = numpy.ones_like(h)
        return hlsa_to_rgba(h, x, ones, ones)


class HLSSaturationColorModel(CylindricalColorModel):
    def __call__(self, h: RealArrayType, x: RealArrayType) -> RealArrayType:
        ones = numpy.ones_like(h)
        return hlsa_to_rgba(h, ones / 2.0, x, ones)


class HLSAlphaColorModel(CylindricalColorModel):
    def __call__(self, h: RealArrayType, x: RealArrayType) -> RealArrayType:
        ones = numpy.ones_like(h)
        return hlsa_to_rgba(h, ones / 2.0, ones, x)


class CylindricalColorModelParameter(Parameter[str], Observer):
    def __init__(self) -> None:
        super().__init__()
        self._chooser = PluginChooser[CylindricalColorModel]()
        self._chooser.register_plugin(
            HSVSaturationColorModel(),
            simple_name='HSV-S',
            display_name='HSV Saturation',
        )
        self._chooser.register_plugin(
            HSVValueColorModel(),
            simple_name='HSV-V',
            display_name='HSV Value',
        )
        self._chooser.register_plugin(
            HSVAlphaColorModel(),
            simple_name='HSV-A',
            display_name='HSV Alpha',
        )
        self._chooser.register_plugin(
            HLSLightnessColorModel(),
            simple_name='HLS-L',
            display_name='HLS Lightness',
        )
        self._chooser.register_plugin(
            HLSSaturationColorModel(),
            simple_name='HLS-S',
            display_name='HLS Saturation',
        )
        self._chooser.register_plugin(
            HLSAlphaColorModel(),
            simple_name='HLS-A',
            display_name='HLS Alpha',
        )
        self.set_value('HSV-V')
        self._chooser.add_observer(self)

    def choices(self) -> Iterator[str]:
        for plugin in self._chooser:
            yield plugin.display_name

    def get_value(self) -> str:
        return self._chooser.get_current_plugin().display_name

    def set_value(self, value: str, *, notify: bool = True) -> None:
        self._chooser.set_current_plugin(value)

    def get_value_as_string(self) -> str:
        return self.get_value()

    def set_value_from_string(self, value: str) -> None:
        self.set_value(value)

    def copy(self) -> Parameter[str]:
        parameter = CylindricalColorModelParameter()
        parameter.set_value(self.get_value())
        return parameter

    def get_plugin(self) -> CylindricalColorModel:
        return self._chooser.get_current_plugin().strategy

    def _update(self, observable: Observable) -> None:
        if observable is self._chooser:
            self.notify_observers()



from __future__ import annotations
from enum import IntEnum

import numpy

from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition

from .builder import ProbePositionsBuilder
from .settings import ProbePositionsSettings


class CartesianProbePositionsVariant(IntEnum):
    RECTANGULAR_RASTER = 0x0
    RECTANGULAR_SNAKE = 0x1
    TRIANGULAR_RASTER = 0x2
    TRIANGULAR_SNAKE = 0x3
    SQUARE_RASTER = 0x4
    SQUARE_SNAKE = 0x5
    HEXAGONAL_RASTER = 0x6
    HEXAGONAL_SNAKE = 0x7

    @property
    def is_snaked(self) -> bool:
        return self.value & 1 != 0

    @property
    def is_triangular(self) -> bool:
        return self.value & 2 != 0

    @property
    def is_equilateral(self) -> bool:
        return self.value & 4 != 0


class CartesianProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(
        self, variant: CartesianProbePositionsVariant, settings: ProbePositionsSettings
    ) -> None:
        super().__init__(settings, variant.name.lower())
        self._variant = variant
        self._settings = settings

        self.num_points_x = settings.num_points_x.copy()
        self._add_parameter('num_points_x', self.num_points_x)

        self.num_points_y = settings.num_points_y.copy()
        self._add_parameter('num_points_y', self.num_points_y)

        self.step_size_x_m = settings.step_size_x_m.copy()
        self._add_parameter('step_size_x_m', self.step_size_x_m)

        self.step_size_y_m = settings.step_size_y_m.copy()
        self._add_parameter('step_size_y_m', self.step_size_y_m)

    def copy(self) -> CartesianProbePositionsBuilder:
        builder = CartesianProbePositionsBuilder(self._variant, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    @property
    def is_equilateral(self) -> bool:
        return self._variant.is_equilateral

    def build(self) -> ProbePositionSequence:
        nx = self.num_points_x.get_value()
        ny = self.num_points_y.get_value()
        dx = self.step_size_x_m.get_value()

        if self._variant.is_equilateral:
            dy = dx

            if self._variant.is_triangular:
                dy *= numpy.sqrt(0.75)
        else:
            dy = self.step_size_y_m.get_value()

        point_list: list[ProbePosition] = list()

        for index in range(nx * ny):
            y, x = divmod(index, nx)

            if self._variant.is_snaked:
                if y & 1:
                    x = nx - 1 - x

            cx = (nx - 1) / 2
            cy = (ny - 1) / 2

            xf = (x - cx) * dx
            yf = (y - cy) * dy

            if self._variant.is_triangular:
                if y & 1:
                    xf += dx / 4
                else:
                    xf -= dx / 4

            point = ProbePosition(
                index=index,
                coordinate_x_m=xf,
                coordinate_y_m=yf,
            )
            point_list.append(point)

        return ProbePositionSequence(point_list)



import logging

from PyQt5.QtCore import Qt, QAbstractItemModel, QModelIndex, QObject
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import (
    QApplication,
    QComboBox,
    QStyle,
    QStyleOptionComboBox,
    QStyleOptionViewItem,
    QStyledItemDelegate,
    QWidget,
)

logger = logging.getLogger(__name__)


class ComboBoxItemDelegate(QStyledItemDelegate):
    # https://wiki.qt.io/Combo_Boxes_in_Item_Views

    def __init__(self, model: QAbstractItemModel, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._model = model
        self._paint_combo_box = False

    def paint(
        self, painter: QPainter | None, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        if painter is None:
            return

        if self._paint_combo_box and index.flags() & Qt.ItemFlag.ItemIsEditable:
            opt = QStyleOptionComboBox()
            opt.rect = option.rect
            opt.currentText = index.data(Qt.ItemDataRole.DisplayRole)
            style = QApplication.style()

            if style is not None:
                style.drawComplexControl(QStyle.ComplexControl.CC_ComboBox, opt, painter)
                style.drawControl(QStyle.ControlElement.CE_ComboBoxLabel, opt, painter)
        else:
            super().paint(painter, option, index)

    def createEditor(  # noqa: N802
        self, parent: QWidget | None, option: QStyleOptionViewItem, index: QModelIndex
    ) -> QWidget:
        combo_box = QComboBox(parent)
        combo_box.activated.connect(self._commit_data_and_close_editor)
        combo_box.setModel(self._model)
        return combo_box

    def setEditorData(self, editor: QWidget | None, index: QModelIndex) -> None:  # noqa: N802
        if editor is None:
            return

        if isinstance(editor, QComboBox):
            current_text = str(index.data(Qt.ItemDataRole.EditRole))
            combo_box_index = editor.findText(current_text)

            if combo_box_index >= 0:
                editor.setCurrentIndex(combo_box_index)

            editor.showPopup()
        else:
            super().setEditorData(editor, index)

    def setModelData(  # noqa: N802
        self, editor: QWidget | None, model: QAbstractItemModel | None, index: QModelIndex
    ) -> None:
        if editor is None or model is None:
            return

        if isinstance(editor, QComboBox):
            model.setData(index, editor.currentText(), Qt.ItemDataRole.EditRole)
        else:
            super().setModelData(editor, model, index)

    def updateEditorGeometry(  # noqa: N802
        self, editor: QWidget | None, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        if editor is None:
            return

        editor.setGeometry(option.rect)

    def _commit_data_and_close_editor(self) -> None:
        editor = self.sender()

        if isinstance(editor, QComboBox):
            self.commitData.emit(editor)
            self.closeEditor.emit(editor)
        else:
            logger.warning('Failed to commit data and close editor! Unexpected editor.')



from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class ProbePositionsSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('ProbePositions')
        self._group.add_observer(self)

        self.builder = self._group.create_string_parameter('Builder', 'rectangular_raster')
        self.file_path = self._group.create_path_parameter('FilePath', Path('/path/to/scan.csv'))
        self.file_type = self._group.create_string_parameter('FileType', 'CSV')

        self.affine00 = self._group.create_real_parameter('Affine00', 1.0)
        self.affine01 = self._group.create_real_parameter('Affine01', 0.0)
        self.affine02 = self._group.create_real_parameter('Affine02', 0.0)
        self.affine10 = self._group.create_real_parameter('Affine10', 0.0)
        self.affine11 = self._group.create_real_parameter('Affine11', 1.0)
        self.affine12 = self._group.create_real_parameter('Affine12', 0.0)
        self.jitter_radius_m = self._group.create_real_parameter(
            'JitterRadiusInMeters', 0.0, minimum=0.0
        )

        self.expand_bbox = self._group.create_boolean_parameter('ExpandBoundingBox', False)
        self.expand_bbox_xmin_m = self._group.create_real_parameter(
            'ExpandedBoundingBoxMinimumXInMeters', -5e-7
        )
        self.expand_bbox_xmax_m = self._group.create_real_parameter(
            'ExpandedBoundingBoxMaximumXInMeters', +5e-7
        )
        self.expand_bbox_ymin_m = self._group.create_real_parameter(
            'ExpandedBoundingBoxMinimumYInMeters', -5e-7
        )
        self.expand_bbox_ymax_m = self._group.create_real_parameter(
            'ExpandedBoundingBoxMaximumYInMeters', +5e-7
        )

        self.num_points_x = self._group.create_integer_parameter('NumberOfPointsX', 10, minimum=0)
        self.num_points_y = self._group.create_integer_parameter('NumberOfPointsY', 10, minimum=0)
        self.step_size_x_m = self._group.create_real_parameter(
            'StepSizeXInMeters', 1e-6, minimum=0.0
        )
        self.step_size_y_m = self._group.create_real_parameter(
            'StepSizeYInMeters', 1e-6, minimum=0.0
        )

        self.radial_step_size_m = self._group.create_real_parameter(
            'RadialStepSizeInMeters', 1e-6, minimum=0.0
        )
        self.num_shells = self._group.create_integer_parameter('NumberOfShells', 5, minimum=0)
        self.num_points_in_first_shell = self._group.create_integer_parameter(
            'NumberOfPointsInFirstShell', 10, minimum=0
        )

        self.amplitude_x_m = self._group.create_real_parameter(
            'AmplitudeXInMeters', 4.5e-6, minimum=0.0
        )
        self.amplitude_y_m = self._group.create_real_parameter(
            'AmplitudeYInMeters', 4.5e-6, minimum=0.0
        )
        self.angular_step_x_turns = self._group.create_real_parameter('AngularStepXInTurns', 0.03)
        self.angular_step_y_turns = self._group.create_real_parameter('AngularStepYInTurns', 0.04)
        self.angular_shift_turns = self._group.create_real_parameter('AngularShiftInTurns', 0.25)

        self.radius_scalar_m = self._group.create_real_parameter(
            'RadiusScalarInMeters', 0.5e-6, minimum=0.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



#!/usr/bin/env python

from pathlib import Path
import argparse
import logging
import sys

from ptychodus.cli import DirectoryType, verify_all_arguments_parsed
from ptychodus.model import ModelCore
import ptychodus

logger = logging.getLogger(__name__)


def main() -> int:
    parser = argparse.ArgumentParser(
        prog=ptychodus.__name__.lower(),
        description=f'{ptychodus.__name__} is a ptychography data pipeline application',
    )
    parser.add_argument(
        '-b',
        '--batch',
        choices=('reconstruct', 'train'),
        help='Run action non-interactively',
    )
    parser.add_argument(
        '-i',
        '--input-directory',
        metavar='INPUT_DIR',
        type=DirectoryType(must_exist=True),
        help='Path to the input data directory (batch mode only)',
    )
    parser.add_argument(
        '--log-level',
        type=int,
        default=logging.INFO,
        help='Set Python logging level.',
    )
    parser.add_argument(
        '-o',
        '--output-directory',
        metavar='OUTPUT_DIR',
        type=DirectoryType(must_exist=False),
        help='Path to the output data directory (batch mode only)',
    )
    parser.add_argument(
        '-s',
        '--settings',
        metavar='SETTINGS_FILE',
        help='Path to the settings file.',
        type=argparse.FileType('r'),
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=ptychodus.VERSION_STRING,
    )

    parsed_args, unparsed_args = parser.parse_known_args()
    settings_file = Path(parsed_args.settings.name) if parsed_args.settings else None

    with ModelCore(settings_file, log_level=parsed_args.log_level) as model:
        if parsed_args.batch is not None:
            verify_all_arguments_parsed(parser, unparsed_args)

            if parsed_args.input_directory is None or parsed_args.output_directory is None:
                parser.error('Batch mode requires input and output arguments!')
                return -1

            return model.batch_mode_execute(
                parsed_args.batch, parsed_args.input_directory, parsed_args.output_directory
            )

        try:
            from PyQt5.QtWidgets import QApplication
        except ModuleNotFoundError:
            logger.warning('PyQt5 not found.')
            return 0

        # QApplication expects the first argument to be the program name
        app = QApplication(sys.argv[:1] + unparsed_args)
        verify_all_arguments_parsed(parser, app.arguments()[1:])

        from ptychodus.view import ViewCore

        view = ViewCore()

        from ptychodus.controller import ControllerCore

        controller = ControllerCore(
            model, view, is_developer_mode_enabled=model.is_developer_mode_enabled
        )
        controller.show_main_window(ptychodus.VERSION_STRING)

        return app.exec()


sys.exit(main())



from pathlib import Path
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent, PixelGeometry
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    DiffractionArray,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry

from .h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder

logger = logging.getLogger(__name__)


class ISNDiffractionFileReader(DiffractionFileReader):
    def __init__(self) -> None:
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        array_list: list[DiffractionArray] = []

        with h5py.File(file_path, 'r') as h5_file:
            ptycho = h5_file['PTYCHO']
            configs = h5_file['configs']
            num_patterns_per_array: list[int] = []

            if isinstance(ptycho, h5py.Group):
                data_path = '/entry/data/data'
                offset = 0

                for name in sorted(ptycho):
                    h5_item = ptycho.get(name, getlink=True)

                    if isinstance(h5_item, h5py.ExternalLink):
                        data_file_path = file_path.parent / h5_item.filename

                        with h5py.File(data_file_path) as h5_data_file:
                            h5_data = h5_data_file[data_path]

                            if isinstance(h5_data, h5py.Dataset):
                                num_patterns, _, _ = h5_data.shape
                                num_patterns_per_array.append(num_patterns)
                            else:
                                raise ValueError(
                                    f'Expected {data_file_path}:{data_path} to be a dataset.'
                                )

                        array = H5DiffractionPatternArray(
                            label=name,
                            indexes=numpy.arange(num_patterns) + offset,
                            file_path=data_file_path,
                            data_path=data_path,
                        )
                        array_list.append(array)
                        offset += num_patterns
                    else:
                        logger.debug(f'Skipping "{name}": not an external link.')
            else:
                raise KeyError('PTYCHO is not a group!')

            if isinstance(configs, h5py.Group):
                detector_distance_m = float(
                    configs['det_dist_mm'][()]
                )  # file units are m; typo in dataset name
                detector_width_px = int(configs['det_size_x'][()])
                detector_height_px = int(configs['det_size_y'][()])
                pixel_width_m = float(configs['pix_size_x'][()])
                pixel_height_m = float(configs['pix_size_y'][()])
                probe_energy_eV = float(configs['photon_energy_eV'][()])  # noqa: N806

                metadata = DiffractionMetadata(
                    num_patterns_per_array=num_patterns_per_array,
                    pattern_dtype=numpy.dtype(numpy.uint32),
                    detector_distance_m=detector_distance_m,
                    detector_extent=ImageExtent(detector_width_px, detector_height_px),
                    detector_pixel_geometry=PixelGeometry(pixel_width_m, pixel_height_m),
                    probe_energy_eV=probe_energy_eV,
                    file_path=file_path,
                )
            else:
                raise KeyError('configs is not a group!')

            contents_tree = self._tree_builder.build(h5_file)

            return SimpleDiffractionDataset(metadata, contents_tree, array_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        ISNDiffractionFileReader(),
        simple_name='APS_ISN',
        display_name='APS 19-ID In-Situ Nanoprobe Files (*.h5 *.hdf5)',
    )



from typing import Any

from PyQt5.QtCore import Qt, QAbstractTableModel, QModelIndex, QObject, QUrl
from PyQt5.QtGui import QColor, QFont

from ...model.globus import GlobusStatusRepository


class GlobusStatusTableModel(QAbstractTableModel):
    def __init__(self, presenter: GlobusStatusRepository, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._presenter = presenter
        self._section_headers = [
            'Label',
            'Start Time',
            'Completion Time',
            'Status',
            'Action',
            'Run ID',
        ]
        self._dt_format = '%Y-%m-%d %H:%M:%S'

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            match orientation:
                case Qt.Orientation.Horizontal:
                    return self._section_headers[section]
                case Qt.Orientation.Vertical:
                    return section

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid():
            flow_run = self._presenter[index.row()]

            if role == Qt.ItemDataRole.DisplayRole:
                match index.column():
                    case 0:
                        return flow_run.label
                    case 1:
                        return flow_run.start_time.strftime(self._dt_format)
                    case 2:
                        if flow_run.completion_time is not None:
                            return flow_run.completion_time.strftime(self._dt_format)
                    case 3:
                        return flow_run.status
                    case 4:
                        return flow_run.action
                    case 5:
                        return flow_run.run_id
            elif index.column() == 5:
                match role:
                    case Qt.ItemDataRole.ToolTipRole:
                        return flow_run.run_url
                    case Qt.ItemDataRole.FontRole:
                        font = QFont()
                        font.setUnderline(True)
                        return font
                    case Qt.ItemDataRole.ForegroundRole:
                        color = QColor(Qt.GlobalColor.blue)
                        return color
                    case Qt.ItemDataRole.UserRole:
                        return QUrl(flow_run.run_url)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._presenter)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._section_headers)



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.propagator import AngularSpectrumPropagator, PropagatorParameters

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class RectangularProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, settings: ProbeSettings) -> None:
        super().__init__(settings, 'rectangular')
        self._settings = settings

        self.width_m = settings.rectangle_width_m.copy()
        self._add_parameter('width_m', self.width_m)

        self.height_m = settings.rectangle_height_m.copy()
        self._add_parameter('height_m', self.height_m)

        # from sample to the focal plane
        self.defocus_distance_m = settings.defocus_distance_m.copy()
        self._add_parameter('defocus_distance_m', self.defocus_distance_m)

    def copy(self) -> RectangularProbeBuilder:
        builder = RectangularProbeBuilder(self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        geometry = geometry_provider.get_probe_geometry()
        coords = self.get_transverse_coordinates(geometry)

        aX_m = numpy.fabs(coords.position_x_m)  # noqa: N806
        rx_m = self.width_m.get_value() / 2.0
        aY_m = numpy.fabs(coords.position_y_m)  # noqa: N806
        ry_m = self.height_m.get_value() / 2.0

        is_inside = numpy.logical_and(aX_m < rx_m, aY_m < ry_m)
        rect = numpy.where(is_inside, 1 + 0j, 0j)

        propagator_parameters = PropagatorParameters(
            wavelength_m=geometry_provider.probe_wavelength_m,
            width_px=rect.shape[-1],
            height_px=rect.shape[-2],
            pixel_width_m=geometry.pixel_width_m,
            pixel_height_m=geometry.pixel_height_m,
            propagation_distance_m=self.defocus_distance_m.get_value(),
        )
        propagator = AngularSpectrumPropagator(propagator_parameters)
        array = propagator.propagate(rect)

        return ProbeSequence(
            array=self.normalize(array),
            opr_weights=None,
            pixel_geometry=geometry.get_pixel_geometry(),
        )



from __future__ import annotations
from decimal import Decimal

from PyQt5.QtCore import QStringListModel

from ptychodus.api.observer import Observable, Observer

from ...model.visualization import VisualizationEngine
from ...view.visualization import VisualizationParametersView


class VisualizationParametersController(Observer):
    def __init__(self, engine: VisualizationEngine, view: VisualizationParametersView) -> None:
        super().__init__()
        self._engine = engine
        self._view = view
        self._renderer_model = QStringListModel()
        self._transformation_model = QStringListModel()
        self._variant_model = QStringListModel()

        view.renderer_combo_box.setModel(self._renderer_model)
        view.transformation_combo_box.setModel(self._transformation_model)
        view.variant_combo_box.setModel(self._variant_model)

        view.min_display_value_line_edit.value_changed.connect(
            lambda value: engine.set_min_display_value(float(value))
        )
        view.max_display_value_line_edit.value_changed.connect(
            lambda value: engine.set_max_display_value(float(value))
        )

        self._sync_model_to_view()
        engine.add_observer(self)

        view.renderer_combo_box.textActivated.connect(engine.set_renderer)
        view.transformation_combo_box.textActivated.connect(engine.set_transformation)
        view.variant_combo_box.textActivated.connect(engine.set_variant)

    def _sync_model_to_view(self) -> None:
        self._view.renderer_combo_box.blockSignals(True)
        self._renderer_model.setStringList([name for name in self._engine.renderers()])
        self._view.renderer_combo_box.setCurrentText(self._engine.get_renderer())
        self._view.renderer_combo_box.blockSignals(False)

        self._view.transformation_combo_box.blockSignals(True)
        self._transformation_model.setStringList([name for name in self._engine.transformations()])
        self._view.transformation_combo_box.setCurrentText(self._engine.get_transformation())
        self._view.transformation_combo_box.blockSignals(False)

        self._view.variant_combo_box.blockSignals(True)
        self._variant_model.setStringList([name for name in self._engine.variants()])
        self._view.variant_combo_box.setCurrentText(self._engine.get_variant())
        self._view.variant_combo_box.blockSignals(False)

        self._view.min_display_value_line_edit.set_value(
            Decimal(repr(self._engine.get_min_display_value()))
        )
        self._view.max_display_value_line_edit.set_value(
            Decimal(repr(self._engine.get_max_display_value()))
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._engine:
            self._sync_model_to_view()



from pathlib import Path
from typing import Final
import logging

import h5py

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
)

logger = logging.getLogger(__name__)


class CSSIPositionFileReader(ProbePositionFileReader):
    ONE_MILLIMETER_M: Final[float] = 1e-3

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            try:
                h5_positions = h5_file['/exchange/motor_pos']
            except KeyError:
                logger.exception('Unable to load scan.')
            else:
                for idx, row in enumerate(h5_positions):
                    point = ProbePosition(
                        idx,
                        row[0] * self.ONE_MILLIMETER_M,
                        row[1] * self.ONE_MILLIMETER_M,
                    )
                    point_list.append(point)

        return ProbePositionSequence(point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        CSSIPositionFileReader(),
        simple_name='APS_CSSI',
        display_name='APS 9-ID CSSI Files (*.h5 *.hdf5)',
    )



from typing import Any, Final

from PyQt5.QtCore import QAbstractListModel, QModelIndex, QObject, Qt
from PyQt5.QtGui import QColor

from ...model.agent import ChatHistory, ChatRole


class AgentMessageListModel(QAbstractListModel):
    DARK_BLUE: Final[QColor] = QColor('#243689')
    LIGHT_BLUE: Final[QColor] = QColor('#0492d2')
    DARK_GREEN: Final[QColor] = QColor('#00894d')
    LIGHT_GREEN: Final[QColor] = QColor('#78ca2a')

    def __init__(self, history: ChatHistory, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._history = history

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid():
            message = self._history[index.row()]

            match role:
                case Qt.ItemDataRole.DisplayRole:
                    return message.content
                case Qt.ItemDataRole.TextAlignmentRole:
                    return (
                        Qt.AlignmentFlag.AlignRight
                        if message.role == ChatRole.USER
                        else Qt.AlignmentFlag.AlignLeft
                    )
                case Qt.ItemDataRole.BackgroundRole:
                    return self.LIGHT_BLUE if message.role == ChatRole.USER else self.LIGHT_GREEN
                case Qt.ItemDataRole.ForegroundRole:
                    return self.DARK_BLUE if message.role == ChatRole.USER else self.DARK_GREEN

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._history)



from collections.abc import Sequence
import logging

import torch
import numpy

from ptychi.api import (
    AffineDegreesOfFreedom,
    BatchingModes,
    Devices,
    Directions,
    Dtypes,
    ImageGradientMethods,
    ImageIntegrationMethods,
    LossFunctions,
    OPRWeightSmoothingMethods,
    ObjectPosOriginCoordsMethods,
    Optimizers,
    OrthogonalizationMethods,
    PatchInterpolationMethods,
    PositionCorrectionTypes,
    PtychographyDataOptions,
)
from ptychi.api.options.base import (
    ForwardModelOptions,
    OPRModeWeightsSmoothingOptions,
    ObjectL1NormConstraintOptions,
    ObjectL2NormConstraintOptions,
    ObjectMultisliceRegularizationOptions,
    ObjectSmoothnessConstraintOptions,
    ObjectTotalVariationOptions,
    OptimizationPlan,
    PositionAffineTransformConstraintOptions,
    PositionCorrectionOptions,
    ProbeCenterConstraintOptions,
    ProbeOrthogonalizeIncoherentModesOptions,
    ProbeOrthogonalizeOPRModesOptions,
    ProbePowerConstraintOptions,
    ProbeSupportConstraintOptions,
    RemoveGridArtifactsOptions,
    RemoveObjectProbeAmbiguityOptions,
    SliceSpacingOptions,
)

from ptychodus.api.common import ComplexArrayType, RealArrayType
from ptychodus.api.object import Object, ObjectGeometry, ObjectPosition
from ptychodus.api.probe import ProbeSequence
from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition
from ptychodus.api.product import LossValue, Product, ProductMetadata
from ptychodus.api.reconstructor import ReconstructInput

from ..diffraction import PatternSizer
from .affine import PtyChiAffineDegreesOfFreedom, PtyChiAffineDegreesOfFreedomBitField
from .settings import (
    PtyChiOPRSettings,
    PtyChiObjectSettings,
    PtyChiProbePositionSettings,
    PtyChiProbeSettings,
    PtyChiSettings,
)


__all__ = ['PtyChiOptionsHelper']

logger = logging.getLogger(__name__)


def create_optimization_plan(start: int, stop: int, stride: int) -> OptimizationPlan:
    return OptimizationPlan(start, None if stop < 0 else stop, stride)


def parse_optimizer(text: str) -> Optimizers:
    try:
        optimizer = Optimizers[text.upper()]
    except KeyError:
        logger.warning(f'Failed to parse optimizer "{text}"!')
        optimizer = Optimizers.SGD

    return optimizer


class PtyChiReconstructorOptionsHelper:
    def __init__(self, settings: PtyChiSettings) -> None:
        self._settings = settings

    @property
    def num_epochs(self) -> int:
        return self._settings.num_epochs.get_value()

    @property
    def num_sync_epochs(self) -> int:
        return self._settings.num_sync_epochs.get_value()

    @property
    def batch_size(self) -> int:
        return self._settings.batch_size.get_value()

    @property
    def batching_mode(self) -> BatchingModes:
        batching_mode_str = self._settings.batching_mode.get_value()

        try:
            return BatchingModes[batching_mode_str.upper()]
        except KeyError:
            logger.warning(f'Failed to parse batching mode "{batching_mode_str}"!')
            return BatchingModes.RANDOM

    @property
    def compact_mode_update_clustering(self) -> bool:
        return self._settings.compact_mode_update_clustering.get_value() > 0

    @property
    def compact_mode_update_clustering_stride(self) -> int:
        return self._settings.compact_mode_update_clustering.get_value()

    @property
    def default_device(self) -> Devices:
        return Devices.GPU if self._settings.use_devices.get_value() else Devices.CPU

    @property
    def default_dtype(self) -> Dtypes:
        return Dtypes.FLOAT64 if self._settings.use_double_precision.get_value() else Dtypes.FLOAT32

    @property
    def use_double_precision_for_fft(self) -> bool:
        return self._settings.use_double_precision_for_fft.get_value()

    @property
    def allow_nondeterministic_algorithms(self) -> bool:
        return self._settings.allow_nondeterministic_algorithms.get_value()

    @property
    def random_seed(self) -> int | None:
        return None  # TODO

    @property
    def displayed_loss_function(self) -> LossFunctions | None:
        return LossFunctions.MSE_SQRT  # TODO

    @property
    def forward_model_options(self) -> ForwardModelOptions:
        return ForwardModelOptions(
            low_memory_mode=self._settings.use_low_memory_mode.get_value(),
            pad_for_shift=self._settings.pad_for_shift.get_value(),
        )


class PtyChiObjectOptionsHelper:
    def __init__(self, settings: PtyChiObjectSettings) -> None:
        self._settings = settings

    @property
    def optimizable(self) -> bool:
        return self._settings.is_optimizable.get_value()

    @property
    def optimization_plan(self) -> OptimizationPlan:
        return create_optimization_plan(
            self._settings.optimization_plan_start.get_value(),
            self._settings.optimization_plan_stop.get_value(),
            self._settings.optimization_plan_stride.get_value(),
        )

    @property
    def optimizer(self) -> Optimizers:
        return parse_optimizer(self._settings.optimizer.get_value())

    @property
    def step_size(self) -> float:
        return self._settings.step_size.get_value()

    @property
    def optimizer_params(self) -> dict:  # TODO
        return dict()

    @property
    def slice_spacing_options(self) -> SliceSpacingOptions:
        optimizer = parse_optimizer(self._settings.optimize_slice_spacing_optimizer.get_value())
        return SliceSpacingOptions(
            optimizable=self._settings.optimize_slice_spacing.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.optimize_slice_spacing_start.get_value(),
                self._settings.optimize_slice_spacing_stop.get_value(),
                self._settings.optimize_slice_spacing_stride.get_value(),
            ),
            optimizer=optimizer,
            step_size=self._settings.optimize_slice_spacing_step_size.get_value(),
        )

    @property
    def l1_norm_constraint(self) -> ObjectL1NormConstraintOptions:
        return ObjectL1NormConstraintOptions(
            enabled=self._settings.constrain_l1_norm.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_l1_norm_start.get_value(),
                self._settings.constrain_l1_norm_stop.get_value(),
                self._settings.constrain_l1_norm_stride.get_value(),
            ),
            weight=self._settings.constrain_l1_norm_weight.get_value(),
        )

    @property
    def l2_norm_constraint(self) -> ObjectL2NormConstraintOptions:
        return ObjectL2NormConstraintOptions(
            enabled=self._settings.constrain_l2_norm.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_l2_norm_start.get_value(),
                self._settings.constrain_l2_norm_stop.get_value(),
                self._settings.constrain_l2_norm_stride.get_value(),
            ),
            weight=self._settings.constrain_l2_norm_weight.get_value(),
        )

    @property
    def smoothness_constraint(self) -> ObjectSmoothnessConstraintOptions:
        return ObjectSmoothnessConstraintOptions(
            enabled=self._settings.constrain_smoothness.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_smoothness_start.get_value(),
                self._settings.constrain_smoothness_stop.get_value(),
                self._settings.constrain_smoothness_stride.get_value(),
            ),
            alpha=self._settings.constrain_smoothness_alpha.get_value(),
        )

    @property
    def total_variation(self) -> ObjectTotalVariationOptions:
        return ObjectTotalVariationOptions(
            enabled=self._settings.constrain_total_variation.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_total_variation_start.get_value(),
                self._settings.constrain_total_variation_stop.get_value(),
                self._settings.constrain_total_variation_stride.get_value(),
            ),
            weight=self._settings.constrain_total_variation_weight.get_value(),
        )

    @property
    def remove_grid_artifacts(self) -> RemoveGridArtifactsOptions:
        direction_str = self._settings.remove_grid_artifacts_direction.get_value()

        try:
            direction = Directions[direction_str.upper()]
        except KeyError:
            logger.warning(f'Failed to parse direction "{direction_str}"!')
            direction = Directions.XY

        return RemoveGridArtifactsOptions(
            enabled=self._settings.remove_grid_artifacts.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.remove_grid_artifacts_start.get_value(),
                self._settings.remove_grid_artifacts_stop.get_value(),
                self._settings.remove_grid_artifacts_stride.get_value(),
            ),
            period_x_m=self._settings.remove_grid_artifacts_period_x_m.get_value(),
            period_y_m=self._settings.remove_grid_artifacts_period_y_m.get_value(),
            window_size=self._settings.remove_grid_artifacts_window_size_px.get_value(),
            direction=direction,
        )

    @property
    def multislice_regularization(self) -> ObjectMultisliceRegularizationOptions:
        unwrap_image_grad_method_str = (
            self._settings.regularize_multislice_unwrap_phase_image_gradient_method.get_value()
        )

        try:
            unwrap_image_grad_method = ImageGradientMethods[unwrap_image_grad_method_str.upper()]
        except KeyError:
            logger.warning(
                f'Failed to parse image gradient method "{unwrap_image_grad_method_str}"!'
            )
            unwrap_image_grad_method = ImageGradientMethods.FOURIER_SHIFT

        unwrap_image_integration_method_str = (
            self._settings.regularize_multislice_unwrap_phase_image_integration_method.get_value()
        )

        try:
            unwrap_image_integration_method = ImageIntegrationMethods[
                unwrap_image_integration_method_str.upper()
            ]
        except KeyError:
            logger.warning(
                f'Failed to parse image integrationient method "{unwrap_image_integration_method_str}"!'
            )
            unwrap_image_integration_method = ImageIntegrationMethods.DECONVOLUTION

        return ObjectMultisliceRegularizationOptions(
            enabled=self._settings.regularize_multislice.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.regularize_multislice_start.get_value(),
                self._settings.regularize_multislice_stop.get_value(),
                self._settings.regularize_multislice_stride.get_value(),
            ),
            weight=self._settings.regularize_multislice_weight.get_value(),
            unwrap_phase=self._settings.regularize_multislice_unwrap_phase.get_value(),
            unwrap_image_grad_method=unwrap_image_grad_method,
            unwrap_image_integration_method=unwrap_image_integration_method,
        )

    @property
    def patch_interpolation_method(self) -> PatchInterpolationMethods:
        method_str = self._settings.patch_interpolator.get_value()

        try:
            return PatchInterpolationMethods[method_str.upper()]
        except KeyError:
            logger.warning(f'Failed to parse patch interpolation method "{method_str}"!')
            return PatchInterpolationMethods.FOURIER

    @property
    def remove_object_probe_ambiguity(self) -> RemoveObjectProbeAmbiguityOptions:
        return RemoveObjectProbeAmbiguityOptions(
            enabled=self._settings.remove_object_probe_ambiguity.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.remove_object_probe_ambiguity_start.get_value(),
                self._settings.remove_object_probe_ambiguity_stop.get_value(),
                self._settings.remove_object_probe_ambiguity_stride.get_value(),
            ),
        )

    @property
    def build_preconditioner_with_all_modes(self) -> bool:
        return self._settings.build_preconditioner_with_all_modes.get_value()

    @property
    def determine_position_origin_coords_by(self) -> ObjectPosOriginCoordsMethods:
        return ObjectPosOriginCoordsMethods.SPECIFIED

    def get_initial_guess(self, object_: Object) -> ComplexArrayType:
        return object_.get_array()

    def get_slice_spacings_m(self, object_: Object) -> RealArrayType | None:
        slice_spacings_m = object_.layer_spacing_m
        return numpy.array(slice_spacings_m) if slice_spacings_m else None

    def get_pixel_size_m(self, object_: Object) -> float:
        pixel_geometry = object_.get_pixel_geometry()
        return pixel_geometry.width_m

    def get_pixel_aspect_ratio(self, object_: Object) -> float:
        pixel_geometry = object_.get_pixel_geometry()
        return pixel_geometry.get_aspect_ratio()

    def get_position_origin_coords(self, object_: Object) -> RealArrayType:
        # TODO return numpy.zeros(2)
        return torch.zeros(2)  # type: ignore


class PtyChiProbeOptionsHelper:
    def __init__(self, settings: PtyChiProbeSettings) -> None:
        self._settings = settings

    @property
    def optimizable(self) -> bool:
        return self._settings.is_optimizable.get_value()

    @property
    def optimization_plan(self) -> OptimizationPlan:
        return create_optimization_plan(
            self._settings.optimization_plan_start.get_value(),
            self._settings.optimization_plan_stop.get_value(),
            self._settings.optimization_plan_stride.get_value(),
        )

    @property
    def optimizer(self) -> Optimizers:
        return parse_optimizer(self._settings.optimizer.get_value())

    @property
    def step_size(self) -> float:
        return self._settings.step_size.get_value()

    @property
    def optimizer_params(self) -> dict:  # TODO
        return dict()

    @property
    def orthogonalize_incoherent_modes(self) -> ProbeOrthogonalizeIncoherentModesOptions:
        method_str = self._settings.orthogonalize_incoherent_modes_method.get_value()

        try:
            method = OrthogonalizationMethods[method_str.upper()]
        except KeyError:
            logger.warning(f'Failed to parse batching mode "{method_str}"!')
            method = OrthogonalizationMethods.GS

        return ProbeOrthogonalizeIncoherentModesOptions(
            enabled=self._settings.orthogonalize_incoherent_modes.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.orthogonalize_incoherent_modes_start.get_value(),
                self._settings.orthogonalize_incoherent_modes_stop.get_value(),
                self._settings.orthogonalize_incoherent_modes_stride.get_value(),
            ),
            method=method,
        )

    @property
    def orthogonalize_opr_modes(self) -> ProbeOrthogonalizeOPRModesOptions:
        return ProbeOrthogonalizeOPRModesOptions(
            enabled=self._settings.orthogonalize_opr_modes.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.orthogonalize_opr_modes_start.get_value(),
                self._settings.orthogonalize_opr_modes_stop.get_value(),
                self._settings.orthogonalize_opr_modes_stride.get_value(),
            ),
        )

    @property
    def support_constraint(self) -> ProbeSupportConstraintOptions:
        return ProbeSupportConstraintOptions(
            enabled=self._settings.constrain_support.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_support_start.get_value(),
                self._settings.constrain_support_stop.get_value(),
                self._settings.constrain_support_stride.get_value(),
            ),
            threshold=self._settings.constrain_support_threshold.get_value(),
        )

    @property
    def center_constraint(self) -> ProbeCenterConstraintOptions:
        return ProbeCenterConstraintOptions(
            enabled=self._settings.constrain_center.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_center_start.get_value(),
                self._settings.constrain_center_stop.get_value(),
                self._settings.constrain_center_stride.get_value(),
            ),
            use_intensity_for_com=self._settings.use_intensity_for_mass_centroid.get_value(),
        )

    @property
    def eigenmode_update_relaxation(self) -> float:
        return self._settings.relax_eigenmode_update.get_value()

    def get_initial_guess(self, probe: ProbeSequence) -> ComplexArrayType:
        return probe.get_array()

    def get_power_constraint(self, metadata: ProductMetadata) -> ProbePowerConstraintOptions:
        return ProbePowerConstraintOptions(
            enabled=self._settings.constrain_probe_power.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_probe_power_start.get_value(),
                self._settings.constrain_probe_power_stop.get_value(),
                self._settings.constrain_probe_power_stride.get_value(),
            ),
            probe_power=metadata.probe_photon_count,
        )


class PtyChiProbePositionOptionsHelper:
    def __init__(self, settings: PtyChiProbePositionSettings) -> None:
        self._settings = settings
        self._affine_dof = PtyChiAffineDegreesOfFreedomBitField(
            settings.constrain_affine_transform_degrees_of_freedom
        )

    @property
    def optimizable(self) -> bool:
        return self._settings.is_optimizable.get_value()

    @property
    def optimization_plan(self) -> OptimizationPlan:
        return create_optimization_plan(
            self._settings.optimization_plan_start.get_value(),
            self._settings.optimization_plan_stop.get_value(),
            self._settings.optimization_plan_stride.get_value(),
        )

    @property
    def optimizer(self) -> Optimizers:
        return parse_optimizer(self._settings.optimizer.get_value())

    @property
    def step_size(self) -> float:
        return self._settings.step_size.get_value()

    @property
    def optimizer_params(self) -> dict:  # TODO
        return dict()

    @property
    def constrain_position_mean(self) -> bool:
        return self._settings.constrain_centroid.get_value()

    @property
    def correction_options(self) -> PositionCorrectionOptions:
        correction_type_str = self._settings.correction_type.get_value()

        try:
            correction_type = PositionCorrectionTypes[correction_type_str.upper()]
        except KeyError:
            logger.warning(f'Failed to parse correction type "{correction_type_str}"!')
            correction_type = PositionCorrectionTypes.GRADIENT

        differentiation_method_str = self._settings.differentiation_method.get_value()

        try:
            differentiation_method = ImageGradientMethods[differentiation_method_str.upper()]
        except KeyError:
            logger.warning(
                f'Failed to parse differentiation method "{differentiation_method_str}"!'
            )
            differentiation_method = ImageGradientMethods.FOURIER_DIFFERENTIATION

        slice_for_correction = (
            self._settings.slice_for_correction.get_value()
            if self._settings.choose_slice_for_correction.get_value()
            else None
        )
        update_magnitude_limit = (
            self._settings.update_magnitude_limit.get_value()
            if self._settings.limit_update_magnitude.get_value()
            else None
        )

        return PositionCorrectionOptions(
            correction_type=correction_type,
            differentiation_method=differentiation_method,
            cross_correlation_scale=self._settings.cross_correlation_scale.get_value(),
            cross_correlation_real_space_width=self._settings.cross_correlation_real_space_width.get_value(),
            cross_correlation_probe_threshold=self._settings.cross_correlation_probe_threshold.get_value(),
            slice_for_correction=slice_for_correction,  # type: ignore
            clip_update_magnitude_by_mad=self._settings.clip_update_magnitude_by_mad.get_value(),
            update_magnitude_limit=update_magnitude_limit,
        )

    @property
    def affine_transform_constraint(self) -> PositionAffineTransformConstraintOptions:
        degrees_of_freedom: list[AffineDegreesOfFreedom] = list()

        if self._affine_dof.is_bit_set(PtyChiAffineDegreesOfFreedom.TRANSLATION):
            degrees_of_freedom.append(AffineDegreesOfFreedom.TRANSLATION)

        if self._affine_dof.is_bit_set(PtyChiAffineDegreesOfFreedom.ROTATION):
            degrees_of_freedom.append(AffineDegreesOfFreedom.ROTATION)

        if self._affine_dof.is_bit_set(PtyChiAffineDegreesOfFreedom.SCALING):
            degrees_of_freedom.append(AffineDegreesOfFreedom.SCALE)

        if self._affine_dof.is_bit_set(PtyChiAffineDegreesOfFreedom.SHEARING):
            degrees_of_freedom.append(AffineDegreesOfFreedom.SHEAR)

        if self._affine_dof.is_bit_set(PtyChiAffineDegreesOfFreedom.ASYMMETRY):
            degrees_of_freedom.append(AffineDegreesOfFreedom.ASYMMETRY)

        return PositionAffineTransformConstraintOptions(
            enabled=self._settings.constrain_affine_transform.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.constrain_affine_transform_start.get_value(),
                self._settings.constrain_affine_transform_stop.get_value(),
                self._settings.constrain_affine_transform_stride.get_value(),
            ),
            degrees_of_freedom=degrees_of_freedom,
            position_weight_update_interval=self._settings.constrain_affine_transform_position_weight_update_interval.get_value(),
            apply_constraint=self._settings.constrain_affine_transform_apply_constraint.get_value(),
            max_expected_error=self._settings.constrain_affine_transform_max_expected_error_px.get_value(),
        )

    def get_positions_px(
        self, scan: ProbePositionSequence, object_geometry: ObjectGeometry
    ) -> tuple[RealArrayType, RealArrayType]:
        position_x_px: list[float] = list()
        position_y_px: list[float] = list()

        for scan_point in scan:
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            position_x_px.append(object_point.coordinate_x_px)
            position_y_px.append(object_point.coordinate_y_px)

        return numpy.array(position_x_px), numpy.array(position_y_px)


class PtyChiOPROptionsHelper:
    def __init__(self, settings: PtyChiOPRSettings) -> None:
        self._settings = settings

    @property
    def optimizable(self) -> bool:
        return self._settings.is_optimizable.get_value()

    @property
    def optimization_plan(self) -> OptimizationPlan:
        return create_optimization_plan(
            self._settings.optimization_plan_start.get_value(),
            self._settings.optimization_plan_stop.get_value(),
            self._settings.optimization_plan_stride.get_value(),
        )

    @property
    def optimizer(self) -> Optimizers:
        return parse_optimizer(self._settings.optimizer.get_value())

    @property
    def step_size(self) -> float:
        return self._settings.step_size.get_value()

    @property
    def optimizer_params(self) -> dict:  # TODO
        return dict()

    @property
    def optimize_eigenmode_weights(self) -> bool:
        return self._settings.optimize_eigenmode_weights.get_value()

    @property
    def optimize_intensity_variation(self) -> bool:
        return self._settings.optimize_intensities.get_value()

    @property
    def smoothing(self) -> OPRModeWeightsSmoothingOptions:
        method_str = self._settings.smoothing_method.get_value()

        try:
            method: OPRWeightSmoothingMethods = OPRWeightSmoothingMethods[method_str.upper()]
        except KeyError:
            logger.debug('OPR weight smoothing method is None.')
            method = OPRWeightSmoothingMethods.MEDIAN

        return OPRModeWeightsSmoothingOptions(
            enabled=self._settings.smooth_mode_weights.get_value(),
            optimization_plan=create_optimization_plan(
                self._settings.smooth_mode_weights_start.get_value(),
                self._settings.smooth_mode_weights_stop.get_value(),
                self._settings.smooth_mode_weights_stride.get_value(),
            ),
            method=method,
            polynomial_degree=self._settings.polynomial_smoothing_degree.get_value(),
        )

    @property
    def update_relaxation(self) -> float:
        return self._settings.relax_update.get_value()

    def get_initial_weights(self, probe: ProbeSequence) -> RealArrayType:
        try:
            return probe.get_opr_weights()
        except ValueError:
            pass

        initial_weights = numpy.zeros((probe.num_coherent_modes))
        initial_weights[0] = 1.0
        return initial_weights


class PtyChiOptionsHelper:
    def __init__(
        self,
        reconstructor_settings: PtyChiSettings,
        object_settings: PtyChiObjectSettings,
        probe_settings: PtyChiProbeSettings,
        probe_position_settings: PtyChiProbePositionSettings,
        opr_settings: PtyChiOPRSettings,
        pattern_sizer: PatternSizer,
    ) -> None:
        self._reconstructor_settings = reconstructor_settings
        self._pattern_sizer = pattern_sizer

        self.reconstructor_helper = PtyChiReconstructorOptionsHelper(reconstructor_settings)
        self.object_helper = PtyChiObjectOptionsHelper(object_settings)
        self.probe_helper = PtyChiProbeOptionsHelper(probe_settings)
        self.probe_position_helper = PtyChiProbePositionOptionsHelper(probe_position_settings)
        self.opr_helper = PtyChiOPROptionsHelper(opr_settings)

    def create_data_options(self, parameters: ReconstructInput) -> PtychographyDataOptions:
        metadata = parameters.product.metadata
        pixel_geometry = self._pattern_sizer.get_processed_pixel_geometry()
        free_space_propagation_distance_m = (
            numpy.inf
            if self._reconstructor_settings.use_far_field_propagation
            else metadata.detector_distance_m
        )
        return PtychographyDataOptions(
            data=parameters.diffraction_patterns,
            free_space_propagation_distance_m=free_space_propagation_distance_m,
            wavelength_m=metadata.probe_wavelength_m,
            fft_shift=self._reconstructor_settings.fft_shift_diffraction_patterns.get_value(),
            detector_pixel_size_m=pixel_geometry.width_m,
            valid_pixel_mask=numpy.logical_not(parameters.bad_pixels),
            save_data_on_device=self._reconstructor_settings.save_data_on_device.get_value(),
        )

    def create_product(
        self,
        product: Product,
        position_x_px: torch.Tensor | numpy.ndarray,
        position_y_px: torch.Tensor | numpy.ndarray,
        probe_array: torch.Tensor | numpy.ndarray,
        object_array: torch.Tensor | numpy.ndarray,
        opr_weights: torch.Tensor | numpy.ndarray,
        losses: Sequence[LossValue],
    ) -> Product:
        object_in = product.object_
        object_out = Object(
            array=numpy.array(object_array),
            layer_spacing_m=object_in.layer_spacing_m,
            pixel_geometry=object_in.get_pixel_geometry(),
            center=object_in.get_center(),
        )

        probe_out = ProbeSequence(
            array=numpy.array(probe_array[0]),
            opr_weights=numpy.array(opr_weights),
            pixel_geometry=product.probes.get_pixel_geometry(),
        )

        corrected_scan_points: list[ProbePosition] = list()
        object_geometry = object_in.get_geometry()

        for uncorrected_point, pos_x_px, pos_y_px in zip(
            product.probe_positions, position_x_px, position_y_px
        ):
            object_point = ObjectPosition(
                index=uncorrected_point.index,
                coordinate_x_px=float(pos_x_px),
                coordinate_y_px=float(pos_y_px),
            )
            scan_point = object_geometry.map_coordinates_object_to_probe(object_point)
            corrected_scan_points.append(scan_point)

        scan_out = ProbePositionSequence(corrected_scan_points)

        return Product(
            metadata=product.metadata,
            probe_positions=scan_out,
            probes=probe_out,
            object_=object_out,
            losses=losses,
        )

    @property
    def num_epochs(self) -> int:
        return self.reconstructor_helper.num_epochs

    @property
    def num_sync_epochs(self) -> int:
        return self.reconstructor_helper.num_sync_epochs



"""Settings registry for reading and writing settings files."""

from __future__ import annotations
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from pathlib import Path
import configparser
import logging

from .observer import Observable
from .parametric import (
    ParameterGroup,
    PathParameter,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class PathPrefixChange:
    """Instruction to replace a path prefix when saving settings (for portability)."""

    find_path_prefix: Path
    replacement_path_prefix: Path


class SettingsRegistry(Observable):
    """Central registry of ParameterGroups that can be saved to and loaded from settings files."""

    def __init__(self) -> None:
        super().__init__()
        self._parameter_group = ParameterGroup()
        self._file_filter_list: list[str] = ['Initialization Files (*.ini)']

    def create_group(self, name: str) -> ParameterGroup:
        return self._parameter_group.create_group(name)

    def __iter__(self) -> Iterator[str]:
        return iter(self._parameter_group.groups())

    def __getitem__(self, name: str) -> ParameterGroup:
        return self._parameter_group.get_group(name)

    def __len__(self) -> int:
        return len(self._parameter_group.groups())

    def get_open_file_filters(self) -> Sequence[str]:
        return self._file_filter_list

    def get_open_file_filter(self) -> str:
        return self._file_filter_list[0]

    def open_settings(self, file_path: Path) -> None:
        config = configparser.ConfigParser(interpolation=None)
        logger.debug(f'Reading settings from "{file_path}"')

        try:
            config.read(file_path)
        except Exception as exc:
            logger.exception(exc)
            return

        # TODO generalize to support nested parameter groups
        for group_name, group in self._parameter_group.groups().items():
            try:
                group_config = config[group_name]
            except KeyError:
                pass
            else:
                for parameter_name, parameter in group.parameters().items():
                    try:
                        value_string = group_config[parameter_name]
                    except KeyError:
                        pass
                    else:
                        parameter.set_value_from_string(value_string)

        self.notify_observers()

    def get_save_file_filters(self) -> Sequence[str]:
        return self._file_filter_list

    def get_save_file_filter(self) -> str:
        return self._file_filter_list[0]

    def save_settings(
        self, file_path: Path, change_path_prefix: PathPrefixChange | None = None
    ) -> None:
        config = configparser.ConfigParser(interpolation=None)
        setattr(config, 'optionxform', lambda option: option)

        for group_name, group in self._parameter_group.groups().items():
            config.add_section(group_name)

            for parameter_name, parameter in group.parameters().items():
                value_string = parameter.get_value_as_string()

                if change_path_prefix and isinstance(parameter, PathParameter):
                    modified_path = parameter.change_path_prefix(
                        change_path_prefix.find_path_prefix,
                        change_path_prefix.replacement_path_prefix,
                    )
                    value_string = str(modified_path)

                config.set(group_name, parameter_name, value_string)

        logger.debug(f'Writing settings to "{file_path}"')

        try:
            with file_path.open(mode='w') as config_file:
                config.write(config_file)
        except Exception as exc:
            logger.exception(exc)
            return



from __future__ import annotations
from collections.abc import Iterator
from typing import Final
import logging
import time

from ptychodus.api.fluorescence import (
    DeconvolutionStrategy,
    ElementMap,
    FluorescenceDataset,
    FluorescenceEnhancingAlgorithm,
    UpscalingStrategy,
)
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.product import Product

from .settings import FluorescenceSettings

logger = logging.getLogger(__name__)

__all__ = [
    'TwoStepFluorescenceEnhancingAlgorithm',
]


class TwoStepFluorescenceEnhancingAlgorithm(FluorescenceEnhancingAlgorithm, Observable, Observer):
    SIMPLE_NAME: Final[str] = 'TwoStep'
    DISPLAY_NAME: Final[str] = 'Upscale and Deconvolve'

    def __init__(
        self,
        settings: FluorescenceSettings,
        upscaling_strategy_chooser: PluginChooser[UpscalingStrategy],
        deconvolution_strategy_chooser: PluginChooser[DeconvolutionStrategy],
    ) -> None:
        super().__init__()
        self._upscaling_strategy_chooser = upscaling_strategy_chooser
        self._deconvolution_strategy_chooser = deconvolution_strategy_chooser

        upscaling_strategy_chooser.synchronize_with_parameter(settings.upscaling_strategy)
        upscaling_strategy_chooser.add_observer(self)

        deconvolution_strategy_chooser.synchronize_with_parameter(settings.deconvolution_strategy)
        deconvolution_strategy_chooser.add_observer(self)

    def enhance(self, dataset: FluorescenceDataset, product: Product) -> FluorescenceDataset:
        upscaler = self._upscaling_strategy_chooser.get_current_plugin().strategy
        deconvolver = self._deconvolution_strategy_chooser.get_current_plugin().strategy
        element_maps: list[ElementMap] = list()

        for emap in dataset.element_maps:
            logger.info(f'Enhancing "{emap.name}"...')
            tic = time.perf_counter()
            emap_upscaled = upscaler(emap, product)
            emap_enhanced = deconvolver(emap_upscaled, product)
            toc = time.perf_counter()
            logger.info(f'Enhanced "{emap.name}" in {toc - tic:.4f} seconds.')

            element_maps.append(emap_enhanced)

        return FluorescenceDataset(
            element_maps=element_maps,
            counts_per_second_path=dataset.counts_per_second_path,
            channel_names_path=dataset.channel_names_path,
        )

    def get_upscaling_strategies(self) -> Iterator[str]:
        for plugin in self._upscaling_strategy_chooser:
            yield plugin.display_name

    def get_upscaling_strategy(self) -> str:
        return self._upscaling_strategy_chooser.get_current_plugin().display_name

    def set_upscaling_strategy(self, name: str) -> None:
        self._upscaling_strategy_chooser.set_current_plugin(name)

    def get_deconvolution_strategies(self) -> Iterator[str]:
        for plugin in self._deconvolution_strategy_chooser:
            yield plugin.display_name

    def get_deconvolution_strategy(self) -> str:
        return self._deconvolution_strategy_chooser.get_current_plugin().display_name

    def set_deconvolution_strategy(self, name: str) -> None:
        self._deconvolution_strategy_chooser.set_current_plugin(name)

    def _update(self, observable: Observable) -> None:
        if observable is self._upscaling_strategy_chooser:
            self.notify_observers()
        elif observable is self._deconvolution_strategy_chooser:
            self.notify_observers()



from __future__ import annotations

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QDialog,
    QFormLayout,
    QGridLayout,
    QLabel,
    QMessageBox,
    QWidget,
)

from ...model.product.probe_positions import (
    CartesianProbePositionsBuilder,
    ConcentricProbePositionsBuilder,
    FromFileProbePositionsBuilder,
    FromMemoryProbePositionsBuilder,
    LissajousProbePositionsBuilder,
    ProbePositionTransform,
    ProbePositionsRepositoryItem,
    SpiralProbePositionsBuilder,
)
from ..parametric import (
    CheckableGroupBoxParameterViewController,
    DecimalLineEditParameterViewController,
    LengthWidgetParameterViewController,
    ParameterViewBuilder,
    ParameterViewController,
)
from ...view.widgets import GroupBoxWithPresets

__all__ = [
    'ProbePositionsEditorViewControllerFactory',
]


class ProbePositionsTransformViewController(ParameterViewController):
    def __init__(self, transform: ProbePositionTransform) -> None:
        super().__init__()
        self._widget = GroupBoxWithPresets('Transformation')

        for index, presets_label in enumerate(transform.labels_for_presets()):
            action = self._widget.presets_menu.addAction(presets_label)
            action.triggered.connect(lambda _, index=index: transform.apply_presets(index))

        self._label_ye = QLabel('y\u2032 =')
        self._label_ye.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self._affine00_view_controller = DecimalLineEditParameterViewController(
            transform.affine00, is_signed=True
        )
        self._label_yp0 = QLabel('y +')
        self._affine01_view_controller = DecimalLineEditParameterViewController(
            transform.affine01, is_signed=True
        )
        self._label_xp0 = QLabel('x +')
        self._affine02_view_controller = LengthWidgetParameterViewController(
            transform.affine02, is_signed=True
        )

        self._label_xe = QLabel('x\u2032 =')
        self._label_xe.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self._affine10_view_controller = DecimalLineEditParameterViewController(
            transform.affine10, is_signed=True
        )
        self._label_yp1 = QLabel('y +')
        self._affine11_view_controller = DecimalLineEditParameterViewController(
            transform.affine11, is_signed=True
        )
        self._label_xp1 = QLabel('x +')
        self._affine12_view_controller = LengthWidgetParameterViewController(
            transform.affine12, is_signed=True
        )

        self._jitter_radius_label = QLabel('Jitter Radius:')
        self._jitter_radius_view_controller = LengthWidgetParameterViewController(
            transform.jitter_radius_m, is_signed=False
        )

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._label_ye, 0, 0)
        layout.addWidget(self._affine00_view_controller.get_widget(), 0, 1)
        layout.addWidget(self._label_yp0, 0, 2)
        layout.addWidget(self._affine01_view_controller.get_widget(), 0, 3)
        layout.addWidget(self._label_xp0, 0, 4)
        layout.addWidget(self._affine02_view_controller.get_widget(), 0, 5)

        layout.addWidget(self._label_xe, 1, 0)
        layout.addWidget(self._affine10_view_controller.get_widget(), 1, 1)
        layout.addWidget(self._label_yp1, 1, 2)
        layout.addWidget(self._affine11_view_controller.get_widget(), 1, 3)
        layout.addWidget(self._label_xp1, 1, 4)
        layout.addWidget(self._affine12_view_controller.get_widget(), 1, 5)

        layout.addWidget(self._jitter_radius_label, 2, 0)
        layout.addWidget(self._jitter_radius_view_controller.get_widget(), 2, 1, 1, 5)
        self._widget.contents.setLayout(layout)

    def get_widget(self) -> QWidget:
        return self._widget


class ScanBoundingBoxViewController(CheckableGroupBoxParameterViewController):
    def __init__(self, item: ProbePositionsRepositoryItem) -> None:
        super().__init__(item.expand_bbox, 'Expand Bounding Box')
        self._xmin_controller = LengthWidgetParameterViewController(
            item.expand_bbox_xmin_m, is_signed=True
        )
        self._xmax_controller = LengthWidgetParameterViewController(
            item.expand_bbox_xmax_m, is_signed=True
        )
        self._ymin_controller = LengthWidgetParameterViewController(
            item.expand_bbox_ymin_m, is_signed=True
        )
        self._ymax_controller = LengthWidgetParameterViewController(
            item.expand_bbox_ymax_m, is_signed=True
        )

        layout = QFormLayout()
        layout.addRow('Minimum X:', self._xmin_controller.get_widget())
        layout.addRow('Maximum X:', self._xmax_controller.get_widget())
        layout.addRow('Minimum Y:', self._ymin_controller.get_widget())
        layout.addRow('Maximum Y:', self._ymax_controller.get_widget())
        self.get_widget().setLayout(layout)


class ProbePositionsEditorViewControllerFactory:
    def _append_common_controls(
        self, dialog_builder: ParameterViewBuilder, item: ProbePositionsRepositoryItem
    ) -> None:
        transform = item.get_transform()

        if transform is not None:
            dialog_builder.add_view_controller_to_bottom(
                ProbePositionsTransformViewController(transform)
            )

        dialog_builder.add_view_controller_to_bottom(ScanBoundingBoxViewController(item))

    def create_editor_dialog(
        self, item_name: str, item: ProbePositionsRepositoryItem, parent: QWidget
    ) -> QDialog:
        probe_positions_builder = item.get_builder()
        builder_name = probe_positions_builder.get_name()
        base_probe_positions_group = 'Base Probe Positions'
        title = f'{item_name} [{builder_name}]'

        if isinstance(probe_positions_builder, CartesianProbePositionsBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_spin_box(
                probe_positions_builder.num_points_x,
                'Number of Points X:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_spin_box(
                probe_positions_builder.num_points_y,
                'Number of Points Y:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_length_widget(
                probe_positions_builder.step_size_x_m,
                'Step Size X:',
                group=base_probe_positions_group,
            )

            if not probe_positions_builder.is_equilateral:
                dialog_builder.add_length_widget(
                    probe_positions_builder.step_size_y_m,
                    'Step Size Y:',
                    group=base_probe_positions_group,
                )

            self._append_common_controls(dialog_builder, item)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_positions_builder, ConcentricProbePositionsBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_spin_box(
                probe_positions_builder.num_shells,
                'Number of Shells:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_spin_box(
                probe_positions_builder.num_points_1st_shell,
                'Number of Points in First Shell:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_length_widget(
                probe_positions_builder.radial_step_size_m,
                'Radial Step Size:',
                group=base_probe_positions_group,
            )
            self._append_common_controls(dialog_builder, item)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_positions_builder, FromFileProbePositionsBuilder):
            dialog_builder = ParameterViewBuilder()
            self._append_common_controls(dialog_builder, item)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_positions_builder, FromMemoryProbePositionsBuilder):
            dialog_builder = ParameterViewBuilder()
            self._append_common_controls(dialog_builder, item)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_positions_builder, SpiralProbePositionsBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_spin_box(
                probe_positions_builder.num_points,
                'Number of Points:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_length_widget(
                probe_positions_builder.radius_scalar_m,
                'Radius Scalar:',
                group=base_probe_positions_group,
            )
            self._append_common_controls(dialog_builder, item)
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(probe_positions_builder, LissajousProbePositionsBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_spin_box(
                probe_positions_builder.num_points,
                'Number of Points:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_length_widget(
                probe_positions_builder.amplitude_x_m,
                'Amplitude X:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_length_widget(
                probe_positions_builder.amplitude_y_m,
                'Amplitude Y:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_angle_widget(
                probe_positions_builder.angular_step_x_turns,
                'Angular Step X:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_angle_widget(
                probe_positions_builder.angular_step_y_turns,
                'Angular Step Y:',
                group=base_probe_positions_group,
            )
            dialog_builder.add_angle_widget(
                probe_positions_builder.angular_shift_turns,
                'Angular Shift:',
                group=base_probe_positions_group,
            )
            self._append_common_controls(dialog_builder, item)
            return dialog_builder.build_dialog(title, parent)

        return QMessageBox(
            QMessageBox.Icon.Information,
            title,
            f'"{builder_name}" has no editable parameters!',
            QMessageBox.StandardButton.Ok,
            parent,
        )



def test_indexing() -> None:
    idx = 0

    for n in range(10):
        print('')

        for m in range(-n, n + 1, 2):
            idx_calc = (n * (n + 2) + m) // 2
            print(f'{n=} {m=:+d} {idx=} {idx_calc=}')
            assert idx == idx_calc
            idx += 1


def test_pyramid() -> None:
    import numpy
    import matplotlib

    matplotlib.use('Agg')
    import matplotlib.colors
    import matplotlib.pyplot as plt

    from ptychodus.model.product.probe.zernike import ZernikePolynomial

    my_dpi = 300
    num_pixels = 256
    max_radial_degree = 6

    Y, X = numpy.mgrid[:num_pixels, :num_pixels]  # noqa: N806
    X = (X - (num_pixels - 1) / 2) / (num_pixels / 2)  # noqa: N806
    Y = (Y - (num_pixels - 1) / 2) / (num_pixels / 2)  # noqa: N806

    distance = numpy.hypot(Y, X)
    angle = numpy.arctan2(Y, X)

    ###

    fig = plt.figure(dpi=my_dpi)
    fig.patch.set_alpha(0.0)
    gs = fig.add_gridspec(max_radial_degree + 1, 2 * (max_radial_degree + 1))

    for radial_degree in range(max_radial_degree):
        for angular_frequency in range(-radial_degree, radial_degree + 1, 2):
            polynomial = ZernikePolynomial(radial_degree, angular_frequency)
            Z = polynomial(distance, angle, undefined_value=numpy.nan)  # noqa: N806

            row = radial_degree
            col = max_radial_degree + angular_frequency

            ax = fig.add_subplot(gs[row : row + 1, col : col + 2])
            ax.pcolormesh(X, Y, Z, norm=matplotlib.colors.CenteredNorm(), cmap='seismic')
            ax.set_aspect('equal')
            ax.set_title(str(polynomial))
            ax.axis('off')

    plt.savefig('zernike_pyramid.png', bbox_inches='tight', dpi=my_dpi)
    plt.close(fig)



from pathlib import Path
from typing import Final
import logging

import h5py
import numpy

from .h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder
from ptychodus.api.geometry import ImageExtent, PixelGeometry
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe import ProbeSequence, ProbeFileReader
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
)
from ptychodus.api.common import ComplexArrayType, ELECTRON_VOLT_J

logger = logging.getLogger(__name__)


class CXIDiffractionFileReader(DiffractionFileReader):
    def __init__(self) -> None:
        self._data_path = '/entry_1/data_1/data'
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            contents_tree = self._tree_builder.build(h5_file)

            data = h5_file[self._data_path]

            if isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape

                detector_extent = ImageExtent(detector_width, detector_height)
                detector_distance_m = float(
                    h5_file['/entry_1/instrument_1/detector_1/distance'][()]
                )
                detector_pixel_geometry = PixelGeometry(
                    float(h5_file['/entry_1/instrument_1/detector_1/x_pixel_size'][()]),
                    float(h5_file['/entry_1/instrument_1/detector_1/y_pixel_size'][()]),
                )
                probe_energy_J = float(h5_file['/entry_1/instrument_1/source_1/energy'][()])  # noqa: N806
                probe_energy_eV = probe_energy_J / ELECTRON_VOLT_J  # noqa: N806

                # TODO load detector mask; zeros are good pixels
                # /entry_1/instrument_1/detector_1/mask Dataset {512, 512}

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_distance_m=detector_distance_m,
                    detector_extent=detector_extent,
                    detector_pixel_geometry=detector_pixel_geometry,
                    probe_energy_eV=probe_energy_eV,
                    file_path=file_path,
                )

                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=file_path,
                    data_path=self._data_path,
                )

                return SimpleDiffractionDataset(metadata, contents_tree, [array])
            else:
                raise ValueError(f'Expected dataset at {self._data_path}, got {type(data)}.')


class CXIPositionFileReader(ProbePositionFileReader):
    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            xyz_m = h5_file['/entry_1/data_1/translation'][()]

            for idx, (x, y, z) in enumerate(xyz_m):
                point = ProbePosition(idx, x, y)
                point_list.append(point)

        return ProbePositionSequence(point_list)


class CXIProbeFileReader(ProbeFileReader):
    def read(self, file_path: Path) -> ProbeSequence:
        array: ComplexArrayType | None = None

        with h5py.File(file_path, 'r') as h5_file:
            array = h5_file['/entry_1/instrument_1/source_1/illumination'][()]

        return ProbeSequence(array=array, opr_weights=None, pixel_geometry=None)


def register_plugins(registry: PluginRegistry) -> None:
    SIMPLE_NAME: Final[str] = 'CXI'  # noqa: N806
    DISPLAY_NAME: Final[str] = 'Coherent X-ray Imaging Files (*.cxi)'  # noqa: N806

    registry.diffraction_file_readers.register_plugin(
        CXIDiffractionFileReader(),
        simple_name=SIMPLE_NAME,
        display_name=DISPLAY_NAME,
    )
    registry.probe_position_file_readers.register_plugin(
        CXIPositionFileReader(),
        simple_name=SIMPLE_NAME,
        display_name=DISPLAY_NAME,
    )
    registry.probe_file_readers.register_plugin(
        CXIProbeFileReader(),
        simple_name=SIMPLE_NAME,
        display_name=DISPLAY_NAME,
    )
